"""
pyndl.corpus
------------

*pyndl.corpus* generates a corpus file (outfile) out of a bunch of gunzipped xml
subtitle files in a directory and all its subdirectories.
"""

import os
import time
import sys
import gzip
import multiprocessing
import xml.etree.ElementTree

__version__ = '0.2.0'

FRAMES_PER_SECOND = 30
PUNCTUATION = tuple(".,:;?!()[]'")


def _parse_time_string(time_string):
    """
    parses string and returns time in seconds.

    """
    # make commas and colons the same symbol and split
    hours, minutes, seconds, frames = time_string.replace(',', ':').split(':')
    return (float(hours) * 60 * 60 +
            float(minutes) * 60 +
            float(seconds) +
            float(frames) / FRAMES_PER_SECOND)


def read_clean_gzfile(gz_file_path, *, break_duration=2.0):
    """
    Generator that opens and reads a gunzipped xml subtitle file, while all
    xml tags and timestamps are removed.

    Parameters
    ----------
    break_duration : float
        defines the amount of time in seconds that need to pass between two
        subtitles in order to start a new paragraph in the resulting corpus.

    Yields
    ------
    line : non empty, cleaned line out of the xml subtitle file

    Raises
    ------
    FileNotFoundError : if file is not there.

    """

    with gzip.open(gz_file_path, "rt", encoding="utf-8-sig") as file_:
        tree = xml.etree.ElementTree.parse(file_)
        root = tree.getroot()

        last_time = 0.0
        for sentence_tag in root.findall('s'):
            # in an s_tag (more or less referring to a 'sentence') there exists
            # time_tags and w_tags (for 'words').

            # join all wordswith spaces in between
            words = []
            for word_tag in sentence_tag.findall('w'):
                text = word_tag.text
                if text in PUNCTUATION:
                    words.append(text)
                elif text is not None:
                    words.extend((' ', text))
                else:
                    raise ValueError("Text content of word tag is None.")
            result = ''.join(words)
            result = result.strip()

            if not result:
                continue

            # Check time and make a new paragraph if needed
            for time_tag in sentence_tag.findall('time'):
                # tag_type is either 'S' or 'E' (start or end)
                tag_type = time_tag.get('id')[-1:]

                current_time = _parse_time_string(time_tag.get('value'))

                # start
                if (tag_type == 'S' and
                        current_time - last_time > break_duration):
                    result = '\n' + result
                # end
                elif tag_type == 'E':
                    last_time = current_time
                elif tag_type == 'S':
                    pass
                else:
                    raise ValueError("tag_type '%s' is not 'S' or 'E'" %
                                     tag_type)

            yield result + "\n"


class JobParseGz():
    # pylint: disable=E0202,missing-docstring

    """
    Stores the persistent information over several jobs and exposes a job
    method that only takes the varying parts as one argument.

    .. note::

        Using a closure is not possible as it is not pickable / serializable.

    """

    def __init__(self, break_duration):
        self.break_duration = break_duration

    def run(self, filename):
        not_found = None
        try:
            lines = list(read_clean_gzfile(filename,
                                           break_duration=self.break_duration))
            lines.append("\n---END.OF.DOCUMENT---\n\n")
        except FileNotFoundError:
            not_found = filename + "\n"
        return (lines, not_found)


def create_corpus_from_gz(directory, outfile, *, n_threads=1, verbose=False):
    """
    Create a corpus file from a set of gunziped (.gz) files in a directory.

    Parameters
    ----------
    directory : str
        use all gz-files in this directory and all subdirectories as input.
    outfile : str
        name of the outfile that will be created.
    n_threads : int
        number of threads to use.
    verbose : bool

    """
    if not os.path.isdir(directory):
        raise OSError("%s does not exist.")
    if os.path.isfile(outfile):
        raise OSError("%s exists. Please <outfile> needs to be new file name."
                      % outfile)

    if verbose:
        print("Walk through '%s' and read in all file names..." % directory)
    gz_files = [os.path.join(root, name)
                for root, dirs, files in os.walk(directory, followlinks=True)
                for name in files
                if name.endswith((".gz",))]
    gz_files.sort()
    if verbose:
        print("Start processing %i files." % len(gz_files))
        start_time = time.time()
    not_founds = list()
    with multiprocessing.Pool(n_threads) as pool:
        with open(outfile, "wt") as result_file:
            progress_counter = 0
            n_files = len(gz_files)
            job = JobParseGz(break_duration=5.0)
            for lines, not_found in pool.imap(job.run, gz_files):
                progress_counter += 1
                if verbose and progress_counter % 1000 == 0:
                    print("%i%% " % (progress_counter / n_files * 100), end="")
                    sys.stdout.flush()

                if lines is not None:
                    result_file.writelines(lines)
                elif not_found is not None:
                    not_founds.append(not_found)
                else:
                    raise NotImplementedError("This should never happend!")
    if verbose:
        duration = time.time() - start_time
        print("\nProcessed %i files. %i files where not found." %
              (len(gz_files), len(not_founds)))
        print("Processing took %.2f seconds (%ih%.2im)." %
              (duration, duration // (60 * 60), duration // 60))

    if not_founds:
        # prevent overwriting files
        file_name = outfile + ".not_found"
        nn = 1
        while not os.path.isfile(file_name):
            nn += 1
            file_name = outfile + ".not_found-" + str(nn)

        with open(file_name, "wt") as not_found_file:
            not_found_file.writelines(not_founds)



import numpy as np
import math
cimport numpy as np
ctypedef np.float64_t dtype_t
cimport cython
from cython.parallel cimport parallel, prange

from ndl_parallel cimport (learn_inplace_binary_to_binary_ptr,
                           learn_inplace_binary_to_real_ptr,
                           learn_inplace_real_to_real_ptr,
                           learn_inplace_real_to_binary_ptr)
from error_codes cimport (ErrorCode,
                          NO_ERROR,
                          MAGIC_NUMBER_DOES_NOT_MATCH,
                          VERSION_NUMBER_DOES_NOT_MATCH,
                          INITIAL_ERROR_CODE,
                          ONLY_ONE_OUTCOME_PER_EVENT,
                          ERROR_CODES)


def learn_inplace_binary_to_binary(binary_file_paths,
                  dtype_t alpha, dtype_t beta1,
                  dtype_t beta2, dtype_t lambda_,
                  np.ndarray[dtype_t, ndim=2] weights,
                  np.ndarray[unsigned int, ndim=1] all_outcomes,
                  unsigned int chunksize,
                  unsigned int n_jobs):

    cdef unsigned int n_all_cues = weights.shape[1]  # number of cues == columns
    cdef unsigned int* all_outcomes_ptr = <unsigned int *> all_outcomes.data
    cdef unsigned int length_all_outcomes = all_outcomes.shape[0]
    cdef char* fname
    cdef unsigned int start_val, end_val, ii, number_parts
    cdef ErrorCode error = INITIAL_ERROR_CODE


  #  cdef String
    # weights muss contigousarray sein und mode=c, siehe:
    #cdef np.ndarray[np.uint32_t, ndim=3, mode = 'c'] np_buff = np.ascontiguousarray(im, dtype = np.uint32)
    cdef dtype_t* weights_ptr = <dtype_t *> weights.data # ueberlegen ob [][] oder ** oder [] oder *

    for binary_file_path in binary_file_paths: #
        filename_byte_string = binary_file_path.encode("UTF-8")
        fname = filename_byte_string

        number_parts = math.ceil(<double> length_all_outcomes / chunksize)

        with nogil, parallel(num_threads=n_jobs):
            for ii in prange(number_parts, schedule="dynamic", chunksize=1):
                start_val = ii * chunksize
                end_val = min(start_val + chunksize, length_all_outcomes)
                if start_val == length_all_outcomes:
                    break
                error = learn_inplace_binary_to_binary_ptr(fname, alpha, beta1,
                                    beta2, lambda_, weights_ptr, n_all_cues,
                                    all_outcomes_ptr, start_val, end_val)
                if error != NO_ERROR:
                    break

    if (error != NO_ERROR):
        raise IOError(f'binary files does not have proper format, error code {error}\n{ERROR_CODES}')


def learn_inplace_binary_to_real(binary_file_paths,
                                 dtype_t eta,
                                 np.ndarray[dtype_t, ndim=2] outcome_vectors,
                                 np.ndarray[dtype_t, ndim=2] weights,
                                 unsigned int chunksize,
                                 unsigned int n_jobs):

    cdef unsigned int n_all_cues = weights.shape[1]  # number of cues == columns
    cdef unsigned int n_outcome_vector_dimensions = outcome_vectors.shape[1]
    cdef char* fname
    cdef unsigned int start_val, end_val, ii, number_parts
    cdef ErrorCode error = INITIAL_ERROR_CODE

    # weights muss contigousarray sein und mode=c, siehe:
    #cdef np.ndarray[np.uint32_t, ndim=3, mode = 'c'] np_buff = np.ascontiguousarray(im, dtype = np.uint32)
    cdef dtype_t* weights_ptr = <dtype_t *> weights.data # ueberlegen ob [][] oder ** oder [] oder *
    cdef dtype_t* outcome_vectors_ptr = <dtype_t *> outcome_vectors.data # ueberlegen ob [][] oder ** oder [] oder *


    for binary_file_path in binary_file_paths:
      filename_byte_string = binary_file_path.encode("UTF-8")
      fname = filename_byte_string

      number_parts = (n_outcome_vector_dimensions // chunksize)
      if n_outcome_vector_dimensions % chunksize != 0:
          number_parts += 1

      with nogil, parallel(num_threads=n_jobs):
        for ii in prange(number_parts, schedule="dynamic", chunksize=1):
          start_val = ii * chunksize
          end_val = min(start_val + chunksize, n_outcome_vector_dimensions)
          if start_val == n_outcome_vector_dimensions:
            break
          error = NO_ERROR
          error = learn_inplace_binary_to_real_ptr(fname,
                                                   eta,
                                                   outcome_vectors_ptr,
                                                   weights_ptr,
                                                   n_all_cues,
                                                   n_outcome_vector_dimensions,
                                                   start_val,
                                                   end_val)
    if (error == ONLY_ONE_OUTCOME_PER_EVENT):
        raise ValueError('error code %i, legal number of outcomes per event is exactly 1')
    if (error == MAGIC_NUMBER_DOES_NOT_MATCH or error == VERSION_NUMBER_DOES_NOT_MATCH):
        raise IOError('binary files does not have proper format, error code %i' % error)


def learn_inplace_real_to_binary(binary_file_paths,
                                 dtype_t beta1,
                                 dtype_t beta2,
                                 dtype_t lambda_,
                                 np.ndarray[dtype_t, ndim=2] cue_vectors,
                                 np.ndarray[dtype_t, ndim=2] weights,
                                 unsigned int chunksize,
                                 unsigned int n_jobs):

    cdef unsigned int n_all_outcomes = weights.shape[0]  # number of outcomes == rows
    cdef unsigned int n_cue_vector_dimensions = weights.shape[1]  # number of cue vector dimensions == columns
    cdef char* fname
    cdef unsigned int start_val, end_val, ii, number_parts
    cdef int error = INITIAL_ERROR_CODE

    # weights muss contigousarray sein und mode=c, siehe:
    #cdef np.ndarray[np.uint32_t, ndim=3, mode = 'c'] np_buff = np.ascontiguousarray(im, dtype = np.uint32)
    cdef dtype_t* weights_ptr = <dtype_t *> weights.data # ueberlegen ob [][] oder ** oder [] oder *
    cdef dtype_t* cue_vectors_ptr = <dtype_t *> cue_vectors.data # ueberlegen ob [][] oder ** oder [] oder *


    for binary_file_path in binary_file_paths:
      filename_byte_string = binary_file_path.encode("UTF-8")
      fname = filename_byte_string

      number_parts = (n_all_outcomes // chunksize)
      if n_all_outcomes % chunksize != 0:
          number_parts += 1

      with nogil, parallel(num_threads=n_jobs):
        for ii in prange(number_parts, schedule="dynamic", chunksize=1):
          start_val = ii * chunksize
          end_val = min(start_val + chunksize, n_all_outcomes)
          if start_val == n_all_outcomes:
            break
          error = NO_ERROR
          error = learn_inplace_real_to_binary_ptr(fname,
                                                   beta1,
                                                   beta2,
                                                   lambda_,
                                                   cue_vectors_ptr,
                                                   weights_ptr,
                                                   n_cue_vector_dimensions,
                                                   start_val,
                                                   end_val)
    if (error == ONLY_ONE_OUTCOME_PER_EVENT):
        raise ValueError('error code %i, legal number of outcomes per event is exactly 1')
    if (error == MAGIC_NUMBER_DOES_NOT_MATCH or error == VERSION_NUMBER_DOES_NOT_MATCH):
        raise IOError('binary files does not have proper format, error code %i' % error)


def learn_inplace_real_to_real(binary_file_paths,
                  dtype_t eta,
                  np.ndarray[dtype_t, ndim=2] cue_vectors,
                  np.ndarray[dtype_t, ndim=2] outcome_vectors,
                  np.ndarray[dtype_t, ndim=2] weights,
                  unsigned int chunksize,
                  unsigned int n_jobs):

    assert weights.shape[1] == cue_vectors.shape[1]
    assert weights.shape[0] == outcome_vectors.shape[1]

    cdef unsigned int n_cue_vector_dimensions = cue_vectors.shape[1]
    cdef unsigned int n_outcome_vector_dimensions = outcome_vectors.shape[1]
    cdef char* fname
    cdef unsigned int start_val, end_val, ii, number_parts
    cdef int error = INITIAL_ERROR_CODE

    # weights muss contigousarray sein und mode=c, siehe:
    #cdef np.ndarray[np.uint32_t, ndim=3, mode = 'c'] np_buff = np.ascontiguousarray(im, dtype = np.uint32)
    cdef dtype_t* weights_ptr = <dtype_t *> weights.data # ueberlegen ob [][] oder ** oder [] oder *
    cdef dtype_t* cue_vectors_ptr = <dtype_t *> cue_vectors.data # ueberlegen ob [][] oder ** oder [] oder *
    cdef dtype_t* outcome_vectors_ptr = <dtype_t *> outcome_vectors.data # ueberlegen ob [][] oder ** oder [] oder *


    for binary_file_path in binary_file_paths:
      filename_byte_string = binary_file_path.encode("UTF-8")
      fname = filename_byte_string

      number_parts = (n_outcome_vector_dimensions // chunksize)
      if n_outcome_vector_dimensions % chunksize != 0:
          number_parts += 1

      with nogil, parallel(num_threads=n_jobs):
        for ii in prange(number_parts, schedule="dynamic", chunksize=1):
          start_val = ii * chunksize
          end_val = min(start_val + chunksize, n_outcome_vector_dimensions)
          if start_val == n_outcome_vector_dimensions:
            break
          error = NO_ERROR
          error = learn_inplace_real_to_real_ptr(fname,
                                                 eta,
                                                 cue_vectors_ptr,
                                                 outcome_vectors_ptr,
                                                 weights_ptr,
                                                 n_cue_vector_dimensions,
                                                 n_outcome_vector_dimensions,
                                                 start_val,
                                                 end_val)
    if (error == ONLY_ONE_OUTCOME_PER_EVENT):
        raise ValueError('error code %i, legal number of outcomes per event is exactly 1')
    if (error == MAGIC_NUMBER_DOES_NOT_MATCH or error == VERSION_NUMBER_DOES_NOT_MATCH):
        raise IOError('binary files does not have proper format, error code %i' % error)



"""
pyndl.io
--------

*pyndl.io* provides functions to create event generators from different
sources in order to use them with *pyndl.ndl* to train NDL models or to save
existing events from a DataFrame or a list to a file.

"""

import gzip
from collections.abc import Iterator, Iterable

import pandas as pd


def events_from_file(event_path, compression="gzip"):
    """
    Yields events for all events in a gzipped event file.

    Parameters
    ----------
    event_path : str
        path to gzipped event file
    compression : str
        indicates whether the events should be read from gunzip
        file or not can be {"gzip" or None}

    Yields
    ------
    cues, outcomes : list, list
        a tuple of two lists containing cues and outcomes

    """
    if compression == "gzip":
        event_file = gzip.open(event_path, 'rt')
    elif compression is None:
        event_file = open(event_path, 'rt')
    else:
        raise ValueError("compression needs to be 'gzip' or None")

    try:
        # skip header
        event_file.readline()
        for line in event_file:
            cues, outcomes = line.strip('\n').split('\t')
            cues = cues.split('_')
            outcomes = outcomes.split('_')
            yield (cues, outcomes)
    finally:
        event_file.close()


def events_to_file(events, file_path, delimiter="\t", compression="gzip",
                   columns=("cues", "outcomes")):
    """
    Writes events to a file

    Parameters
    ----------
    events : pandas.DataFrame or Iterator or Iterable
        a pandas DataFrame with one event per row and one colum with the cues
        and one column with the outcomes or a list of cues and outcomes as strings
        or a list of a list of cues and a list of outcomes which should be written
        to a file
    file_path: str
        path to where the file should be saved
    delimiter: str
        Seperator which should be used. Default ist a tab
    compression : str
        indicates whether the events should be read from gunzip
        file or not can be {"gzip" or None}
    columns: tuple
        a tuple of column names

    """
    if isinstance(events, pd.DataFrame):
        events = events_from_dataframe(events)
    elif isinstance(events, (Iterator, Iterable)):
        events = events_from_list(events)
    else:
        raise ValueError("events should either be a pd.DataFrame or an Iterator or an Iterable.")

    if compression == "gzip":
        out_file = gzip.open(file_path, 'wt')
    elif compression is None:
        out_file = open(file_path, 'wt')
    else:
        raise ValueError("compression needs to be 'gzip' or None")

    try:
        out_file.write("{}\n".format(delimiter.join(columns)))

        for cues, outcomes in events:
            if isinstance(cues, list) and isinstance(outcomes, list):
                line = "{}{}{}\n".format("_".join(cues),
                                         delimiter,
                                         "_".join(outcomes))
            elif isinstance(cues, str) and isinstance(outcomes, str):
                line = "{}{}{}\n".format(cues, delimiter, outcomes)
            else:
                raise ValueError("cues and outcomes should either be a list or a string.")
            out_file.write(line)
    finally:
        out_file.close()


def events_from_dataframe(df, columns=("cues", "outcomes")):
    """
    Yields events for all events in a pandas dataframe.

    Parameters
    ----------
    df : pandas.DataFrame
        a pandas DataFrame with one event per row and one colum with the cues
        and one column with the outcomes.
    columns : tuple
        a tuple of column names

    Yields
    ------
    cues, outcomes : list, list
        a tuple of two lists containing cues and outcomes

    """
    for _, row in df.iterrows():
        cues, outcomes = row[list(columns)]
        cues = cues.split('_')
        outcomes = outcomes.split('_')
        yield (cues, outcomes)


def events_from_list(lst):
    """
    Yields events for all events in a list.

    Parameters
    ----------
    lst : list of list of str or list of str
        a list either containing a list of cues as strings and a list of outcomes
        as strings or a list containing a cue and an outcome string, where cues
        respectively outcomes are seperated by an undescore

    Yields
    ------
    cues, outcomes : list, list
        a tuple of two lists containing cues and outcomes

    """
    for cues, outcomes in lst:
        if isinstance(cues, str):
            cues = cues.split('_')
        if isinstance(outcomes, str):
            outcomes = outcomes.split('_')
        yield (cues, outcomes)



#!/usr/bin/env python3

# pylint: disable=C0111

import os

from pyndl import corpus

TEST_ROOT = os.path.dirname(__file__)


def test_read_clean_gzfile():
    file_name = os.path.join(TEST_ROOT, 'resources/xml_gz_corpus/2017/0001/1.xml.gz')

    lines = list(corpus.read_clean_gzfile(file_name, break_duration=5.0))
    assert len(lines) == 3


def test_create_corpus_from_gz():
    corpus_file = os.path.join(TEST_ROOT, 'temp/xml_gz_corpus.txt')
    resource_file = os.path.join(TEST_ROOT, 'resources/xml_gz_corpus')
    reference_file = os.path.join(TEST_ROOT, 'reference/xml_gz_corpus.txt')
    corpus.create_corpus_from_gz(resource_file, corpus_file, n_threads=1, verbose=False)
    os.remove(corpus_file)
    corpus.create_corpus_from_gz(resource_file, corpus_file, n_threads=2, verbose=True)
    with open(corpus_file, "rt") as new_file:
        lines_new = new_file.readlines()
    with open(reference_file, "rt") as reference:
        lines_reference = reference.readlines()
    assert lines_new == lines_reference
    os.remove(corpus_file)



#!/usr/bin/env python3

# pylint: disable=C0111

import re
from io import StringIO
from contextlib import redirect_stdout

import pyndl


def test_sysinfo():
    out = StringIO()
    with redirect_stdout(out):
        pyndl.sysinfo()
    out = out.getvalue()

    pattern = re.compile(r"[a-zA-Z0-9_\. ]*\n[\=]*\n+([a-zA-Z0-9_ ]*\n[\-]*\n"
                         r"([a-zA-Z0-9_ ]*: [a-zA-Z0-9_\.\-/ ]*\n+)+)+")
    assert pattern.match(out)



'''
Configuration for py.test-3.

'''

import pytest


def pytest_addoption(parser):
    parser.addoption("--runslow", action="store_true",
                     help="run slow tests")
    parser.addoption("--no-linux", action="store_true",
                     help="run without linux tests")


def pytest_runtest_setup(item):
    if 'runslow' in item.keywords and not item.config.getoption("--runslow"):
        pytest.skip("need --runslow option to run this test")

    if 'nolinux' in item.keywords and item.config.getoption("--no-linux"):
        pytest.skip("run without --no-linux option to run this test")


def pytest_configure(config):
    config.addinivalue_line(
        "markers", 'runslow: marks tests which run very slow (deselect with -m "not runslow")'
    )
    config.addinivalue_line(
        "markers", 'nolinux: marks tests which just run on linux (deselect with -m "not nolinux")'
    )



"""
Pyndl - Naive Discriminative Learning in Python
===============================================

*pyndl* is an implementation of Naive Discriminative Learning in Python. It was
created to analyse huge amounts of text file corpora. Especially, it allows to
efficiently apply the Rescorla-Wagner learning rule to these corpora.

"""

import os
import sys
import multiprocessing as mp
try:
    from importlib.metadata import requires
except ModuleNotFoundError:  # python 3.7 and before
    requires = None
try:
    from packaging.requirements import Requirement
except ModuleNotFoundError:  # this should only happend during setup phase
    Requirement = None


__author__ = ('Konstantin Sering, Marc Weitz, '
              'David-Elias Künstle, Lennard Schneider, '
              'Elnaz Shafaei-Bajestan')
__author_email__ = 'konstantin.sering@uni-tuebingen.de'
__version__ = '0.8.1'
__license__ = 'MIT'
__description__ = ('Naive discriminative learning implements learning and '
                   'classification models based on the Rescorla-Wagner '
                   'equations.')
__classifiers__ = [
    'Development Status :: 4 - Beta',
    'Environment :: Console',
    'Intended Audience :: Science/Research',
    'License :: OSI Approved :: MIT License',
    'Operating System :: POSIX :: Linux',
    'Operating System :: MacOS',
    'Programming Language :: Python',
    'Programming Language :: Python :: 3.7',
    'Programming Language :: Python :: 3.8',
    'Programming Language :: Python :: 3.9',
    'Programming Language :: Python :: 3 :: Only',
    'Topic :: Scientific/Engineering',
    'Topic :: Scientific/Engineering :: Artificial Intelligence',
    'Topic :: Scientific/Engineering :: Information Analysis',
    ]


def sysinfo():
    """
    Prints system the dependency information
    """
    if requires:
        dependencies = [Requirement(req).name for req in requires('pyndl')
                        if not Requirement(req).marker]

    header = ("Pyndl Information\n"
              "=================\n\n")

    general = ("General Information\n"
               "-------------------\n"
               "Python version: {}\n"
               "Pyndl version: {}\n\n").format(sys.version.split()[0], __version__)

    uname = os.uname()
    osinfo = ("Operating System\n"
              "----------------\n"
              "OS: {s.sysname} {s.machine}\n"
              "Kernel: {s.release}\n"
              "CPU: {cpu_count}\n").format(s=uname, cpu_count=mp.cpu_count())

    if uname.sysname == "Linux":
        _, *lines = os.popen("free -m").readlines()
        for identifier in ("Mem:", "Swap:"):
            memory = [line for line in lines if identifier in line]
            if len(memory) > 0:
                _, total, used, *_ = memory[0].split()
            else:
                total, used = '?', '?'
            osinfo += "{} {}MiB/{}MiB\n".format(identifier, used, total)

    osinfo += "\n"

    deps = ("Dependencies\n"
            "------------\n")

    if requires:
        deps += "\n".join("{pkg.__name__}: {pkg.__version__}".format(pkg=__import__(dep))
                          for dep in dependencies)
    else:
        deps = 'You need Python 3.8 or higher to show dependencies.'

    print(header + general + osinfo + deps)



"""
Numpy's mypy stub. Only type declarations for ndarray, the scalar hierarchy and array creation
methods are provided.
"""

from typing import (Any, Callable, Dict, Generic, Iterator, List, Optional, Sequence, Tuple, Type,
                    TypeVar, Union)

class dtype: ...
_dtype = dtype


class flagsobj:
    """numpy.flagsobj"""
    aligned = None       # type: bool
    behaved = None       # type: bool
    c_contiguous = None  # type: bool
    carray = None        # type: bool
    contiguous = None    # type: bool
    f_contiguous = None  # type: bool
    farray = None        # type: bool
    fnc = None           # type: bool
    forc = None          # type: bool
    fortran = None       # type: bool
    owndata = None       # type: bool
    updateifcopy = None  # type: bool
    writeable = None     # type: bool
    def __getitem__(self, item: str) -> bool: ...
    def __setitem__(self, item: str, value: bool) -> None: ...

#
# Type variables. _T wasn't used to avoid confusions with ndarray's "T" attribute.
#

_S = TypeVar('_S')
_U = TypeVar('_U')
_V = TypeVar('_V')

#
# Auxiliary types
#

ShapeType = Union[int, Tuple[int, ...]]
AxesType = Union[int, Tuple[int, ...]]
OrderType = Union[str, Sequence[str]]
DtypeType = Union[dtype, type]

class flatiter(Generic[_S], Iterator[_S]):
    coords = ...  # type: ShapeType
    def copy(self) -> flatiter[_S]: ...

class _ArrayLike(Generic[_S]):
    """
    "array-like" interface that both numpy.ndarray and all scalars (descendants of numpy.generic)
    implement this interface.
    """
    #
    # Array-like structures attributes
    #
    T = None         # type: _ArrayLike[_S]
    data = None      # type: Any
    dtype = None     # type: _dtype
    flags = None     # type: flagsobj
    flat = None      # type: flatiter[_ArrayLike[_S]]
    imag = None      # type: _ArrayLike[_S]
    real = None      # type: _ArrayLike[_S]
    size = None      # type: int
    itemsize = None  # type: int
    nbytes = None    # type: int
    ndim = None      # type: int
    shape = None     # type: Tuple[int, ...]
    strides = None   # type: Tuple[int, ...]
    base = None      # type: Optional[_ArrayLike[_S]]

    #
    # Array-like methods
    #

    # Once this issue https://github.com/python/mypy/issues/1907 is resolved, most methods that
    # have an 'out' argument, will be implemented using overload instead of with a Union
    # result. mypy is smart enough to assign the proper type (_ArrayLike[_U]) when out is present
    # but it falls back to the union when it's not.
    def all(self, axis: AxesType=None, out: '_ArrayLike[_U]'=None,
            keepdims: bool=False) -> Union['_ArrayLike[_U]', '_ArrayLike[bool]']: ...

    def any(self, axis: AxesType=None, out: '_ArrayLike[_U]'=None,
            keepdims: bool=False) -> Union['_ArrayLike[_U]', '_ArrayLike[bool]']: ...

    def argmax(self, axis: int=None,
               out: '_ArrayLike[_U]'=None) -> Union['_ArrayLike[_U]', '_ArrayLike[int]']: ...

    def argmin(self, axis: int=None,
               out: '_ArrayLike[_U]'=None) -> Union['_ArrayLike[_U]', '_ArrayLike[int]']: ...

    def argpartition(self, kth: Union[int, Sequence[int]], axis: Optional[int]=-1,
                     kind: str='introselect', order: OrderType=None) -> '_ArrayLike[int]': ...

    def argsort(self, axis: int=None, kind: str='quicksort',
                order: OrderType=None) -> '_ArrayLike[int]': ...

    def astype(self, dtype: Any, order: str='K', casting: str='unsafe', subok: bool=True,
               copy: bool=False) -> '_ArrayLike[Any]': ...

    def byteswap(self, inplace: bool=False) -> '_ArrayLike[_S]': ...

    def choose(self, choices:Sequence['_ArrayLike[_V]'], out: '_ArrayLike[_U]'=None,
               mode: str='raise') ->  Union['_ArrayLike[_U]', '_ArrayLike[_V]']: ...

    def clip(self, a_min: Any, a_max: Any,
             out: '_ArrayLike[_U]'=None) -> Union['_ArrayLike[_S]', '_ArrayLike[_U]']: ...

    def compress(self, condition: Sequence[bool], axis: int=None,
                 out: '_ArrayLike[_U]'=None) -> Union['_ArrayLike[_S]', '_ArrayLike[_U]']: ...

    def conj(self) -> '_ArrayLike[_S]': ...

    def conjugate(self) -> '_ArrayLike[_S]': ...

    def copy(self, order: str='C') -> '_ArrayLike[_S]': ...

    def cumprod(self, axis: int=None, dtype: Any=None,
                out: '_ArrayLike[Any]'=None) -> '_ArrayLike[Any]': ...

    def cumsum(self, axis: int=None, dtype: DtypeType=None,
                out: '_ArrayLike[Any]'=None) -> '_ArrayLike[Any]': ...

    def diagonal(self, offset: int=0, axis1: int=0, axis2: int=1) -> '_ArrayLike[_S]': ...

    def dot(self, b: '_ArrayLike[Any]', out: '_ArrayLike[Any]'=None) -> '_ArrayLike[Any]': ...

    def dump(self, file: str) -> None: ...

    def dumps(self) -> str: ...

    def fill(self, value: _S) -> None: ...

    def flatten(self, order: str='C') -> '_ArrayLike[_S]': ...

    def getfield(self, dtype: DtypeType, offset: int=0) -> '_ArrayLike[Any]': ...

    def item(self, args: AxesType) -> generic: ...

    def itemset(self, arg0: Union[int, Tuple[int, ...]], arg1: Any=None) -> None: ...

    def max(self, axis: AxesType=None,
            out: '_ArrayLike[_U]'=None) -> Union['_ArrayLike[_S]', '_ArrayLike[_U]']: ...

    def mean(self, axis: AxesType=None, dtype: Any=None,
             out: '_ArrayLike[_U]'=None, keepdims: bool=False) -> '_ArrayLike[floating]': ...

    def min(self, axis: AxesType=None,
            out: '_ArrayLike[_U]'=None) -> Union['_ArrayLike[_S]', '_ArrayLike[_U]']: ...

    def newbyteorder(self, new_order: str='S') -> '_ArrayLike[_S]': ...

    def nonzero(self) -> '_ArrayLike[int]': ...

    def partition(self, kth: AxesType, axis: int=-1, kind: str='introselect',
                  order: OrderType=None) -> None: ...

    def prod(self, axis: AxesType=None, dtype: DtypeType=None,
             out: '_ArrayLike[_U]'=None, keepdims: bool=False) -> '_ArrayLike[Any]': ...

    def ptp(self, axis: int=None,
            out: '_ArrayLike[_U]'=None) -> Union['_ArrayLike[_S]', '_ArrayLike[_U]']: ...

    def put(self, ind: '_ArrayLike[int]', v: '_ArrayLike[_S]', mode: str='raise') -> None: ...

    def ravel(self, order: str='C') -> '_ArrayLike[_S]': ...

    def repeat(self, repeats: Union[int, Sequence[int]],
               axis: int=None) -> '_ArrayLike[_S]': ...

    def reshape(self, newshape: ShapeType,
                order: str='C') -> '_ArrayLike[_S]': ...

    def resize(self, new_shape: ShapeType, refcheck: bool=True) -> None: ...

    def round(self, decimals: int=0,
              out: '_ArrayLike[_U]'=None) -> Union['_ArrayLike[_S]', '_ArrayLike[_U]']: ...

    def searchsorted(self, v: Union[_S, '_ArrayLike[_S]'], side: str='left',
                     sorter: '_ArrayLike[int]'=None) -> '_ArrayLike[int]': ...

    def setfield(self, val: Any, dtype: DtypeType, offset: int=0) -> None: ...

    def setflags(self, write: bool=None, align: bool=None,
                 uic: bool=None) -> None: ...

    def sort(self, axis: int=-1, kind: str='quicksort', order: OrderType=None) -> None: ...

    def squeeze(self, axis: AxesType=None) -> '_ArrayLike[_S]': ...

    def std(self, axis: AxesType=None, dtype: DtypeType=None,
            out: '_ArrayLike[_U]'=None, ddof: int=0, keepdims: bool=False) -> '_ArrayLike[floating]': ...

    def sum(self, axis: AxesType=None, dtype: DtypeType=None,
            out: '_ArrayLike[_U]'=None,
            keepdims: bool=False) -> '_ArrayLike[Any]': ...

    def swapaxes(self, axis1: int, axis2: int) -> '_ArrayLike[_S]': ...

    def take(self, indices: Sequence[int], axis: int=None,
             out: '_ArrayLike[_U]'=None,
             mode: str='raise') -> Union['_ArrayLike[_S]', '_ArrayLike[_U]']: ...

    def tobytes(self, order: str='C') -> bytes: ...

    def tofile(self, fid: object, sep: str='',  # TODO fix fid definition (There's a bug in mypy io's namespace https://github.com/python/mypy/issues/1462)
               format: str='%s') -> None: ...

    def tolist(self) -> List[Any]: ...

    def tostring(self, order: str='C') -> bytes: ...

    def trace(self, offset: int=0, axis1: int=0, axis2: int=1,
              dtype: DtypeType=None, out: '_ArrayLike[_U]'=None) -> '_ArrayLike[Any]': ...

    def transpose(self, axes: Optional[AxesType]) -> '_ArrayLike[_S]': ...

    def var(self, axis: AxesType=None, dtype: DtypeType=None,
            out: '_ArrayLike[_U]'=None, ddof: int=0, keepdims: bool=False) -> '_ArrayLike[Any]': ...

    def view(self, dtype: Union[DtypeType, Type['ndarray']]=None,
             type: type=None) -> '_ArrayLike[Any]': ...

    #
    # Magic methods
    #

    def __abs__(self) -> '_ArrayLike[_S]': ...

    def __add__(self, value: object) -> '_ArrayLike[Any]': ...

    def __and__(self, value: object) -> '_ArrayLike[int]': ...

    def __array__(self, dtype: DtypeType=None) -> '_ArrayLike[Any]': ...

    def __array_prepare__(self, context: object=None) -> '_ArrayLike[Any]': ...

    def __array_wrap__(self, context: object=None) -> '_ArrayLike[Any]': ...

    def __bool__(self) -> bool: ...

    def __complex__(self) -> complex: ...

    def __contains__(self, key: object) -> bool: ...

    def __copy__(self) -> '_ArrayLike[_S]': ...

    def __deepcopy__(self) -> '_ArrayLike[_S]': ...

    def __delattr__(self, name: str) -> None: ...

    def __delitem__(self, key: str) -> None: ...

    def __dir__(self) -> List[str]: ...

    def __divmod__(self, value: object) -> Tuple['_ArrayLike[int]', '_ArrayLike[float]']: ...

    def __eq__(self, value: object) -> '_ArrayLike[bool]': ...  # type: ignore

    def __float__(self) -> float: ...

    def __floordiv__(self, value: object) -> '_ArrayLike[int]': ...

    def __ge__(self, value: object) -> '_ArrayLike[bool]': ...

    def __getattribute__(self, name: str) -> Any: ...

    def __getitem__(self, key: Any) -> '_ArrayLike[_S]': ...

    def __gt__(self, value: object) -> '_ArrayLike[bool]': ...

    def __iadd__(self, value: object) -> None: ...

    def __iand__(self, value: object) -> None: ...

    def __ifloordiv__(self, value: object) -> None: ...

    def __ilshift__(self, value: object) -> None: ...

    def __imatmul__(self, value: '_ArrayLike[Any]') -> None: ...

    def __imod__(self, value: object) -> None: ...

    def __imul__(self, value: object) -> None: ...

    def __index__(self) -> int: ...

    def __int__(self) -> int: ...

    def __invert__(self) -> '_ArrayLike[_S]': ...

    def __ior__(self, value: object) -> None: ...

    def __ipow__(self, value: object) -> None: ...

    def __irshift__(self, value: object) -> None: ...

    def __isub__(self, value: object) -> None: ...

    def __iter__(self) -> Iterator['_ArrayLike[_S]']: ...

    def __itruediv__(sel, value: object) -> None: ...

    def __ixor__(self, value: object) -> None: ...

    def __le__(self, value: object) -> '_ArrayLike[bool]': ...

    def __len__(self) -> int: ...

    def __lshift__(self, value: object) -> '_ArrayLike[_S]': ...

    def __lt__(self, value: object) -> '_ArrayLike[bool]': ...

    def __matmul__(self, value: '_ArrayLike[Any]') -> '_ArrayLike[Any]': ...

    def __mod__(self, value: object) -> '_ArrayLike[_S]': ...

    def __mul__(self, value: object) -> '_ArrayLike[Any]': ...

    def __ne__(self, value: object) -> '_ArrayLike[bool]': ...  # type: ignore

    def __neg__(self) -> '_ArrayLike[_S]': ...

    def __or__(self, value: object) -> '_ArrayLike[_S]': ...

    def __pos__(self) -> '_ArrayLike[_S]': ...

    def __pow__(self, value: object) -> '_ArrayLike[Any]': ...

    def __radd__(self, value: object) -> '_ArrayLike[Any]': ...

    def __rand__(self, value: object) -> '_ArrayLike[_S]': ...

    def __rdivmod__(self, value: object) -> Tuple['_ArrayLike[int]', '_ArrayLike[float]']: ...

    def __rfloordiv__(self, value: object) -> '_ArrayLike[Any]': ...

    def __rlshift__(self, value: object) -> '_ArrayLike[Any]': ...

    def __rmatmul__(self, value: object) -> '_ArrayLike[Any]': ...

    def __rmod__(self, value: object) -> '_ArrayLike[Any]': ...

    def __rmul__(self, value: object) -> '_ArrayLike[Any]': ...

    def __ror__(self, value: object) -> '_ArrayLike[_S]': ...

    def __rpow__(self, value: object) -> '_ArrayLike[Any]': ...

    def __rrshift__(self, value: object) -> '_ArrayLike[Any]': ...

    def __rshift__(self, value: object) -> '_ArrayLike[Any]': ...

    def __rsub__(self, value: object) -> '_ArrayLike[Any]': ...

    def __rtruediv__(self, value: object) -> '_ArrayLike[Any]': ...

    def __rxor__(self, value: object) -> '_ArrayLike[_S]': ...

    def __setattr__(self, name: str, value: Any) -> None: ...

    def __setitem__(self, key: Any, value: Any) -> None: ...

    def __str__(self) -> str: ...

    def __sub__(self, value: object) -> '_ArrayLike[Any]': ...

    def __truediv__(sel, value: object) -> '_ArrayLike[Any]': ...

    def __xor__(self, value: object) -> '_ArrayLike[_S]': ...

#
# numpy's scalar hierarchy (http://docs.scipy.org/doc/numpy/reference/arrays.scalars.html#scalars)
#

class generic(_ArrayLike[_S], Generic[_S]): ...
class bool_(generic[bool]): ...
bool8 = bool_
class object_(generic[Any]): ...
class number(generic[_S], Generic[_S]): ...
class integer(number[int]): ...
class signedinteger(integer): ...
class byte(signedinteger): ...
class short(signedinteger): ...
class intc(signedinteger): ...
class int_(signedinteger): ...
class longlong(signedinteger): ...
class int8(signedinteger): ...
class int16(signedinteger): ...
class int32(signedinteger): ...
class int64(signedinteger): ...
class unsignedinteger(integer): ...
class ubyte(unsignedinteger): ...
class ushort(unsignedinteger): ...
class uintc(unsignedinteger): ...
class uint(unsignedinteger): ...
class ulonglong(unsignedinteger): ...
class uint8(signedinteger): ...
class uint16(signedinteger): ...
class uint32(signedinteger): ...
class uint64(signedinteger): ...
class inexact(number[float]): ...
class floating(inexact): ...
class half(floating): ...
class single(floating): ...
class float_(floating): ...
class longfloat_(floating): ...
class float16(floating): ...
class float32(floating): ...
class float64(floating): ...
class float128(floating): ...
class complextfloating(inexact): ...
class csingle(complextfloating): ...
class complex_(complextfloating): ...
class clongfloat(complextfloating): ...
class complex64(complextfloating): ...
class complex128(complextfloating): ...
class complex256(complextfloating): ...
class flexible(generic[_S], Generic[_S]): ...
class character(flexible[str]): ...
class str_(character): ...
class unicode_(character): ...
class void(flexible[None]): ...

class ndarray(_ArrayLike[_S], Generic[_S]):
    """numpy.ndarray"""
    ctypes = None    # type: Any  # TODO Implement ctypes type hint

    # TODO Need to find a way to restrict buffer type
    def __init__(self, shape: Tuple[int, ...], dtype: DtypeType=None,
                 buffer: Any=None, offset: int=None,
                 strides: Tuple[int, ...]=None, order: str=None) -> None: ...

#
# Array creation routines
#

def array(object: Any, dtype: Any=None, copy: bool=True,
          order: str=None, subok: bool=False,
          ndmin: int=0) -> ndarray[Any]: ...
def asarray(a: Any, dtype: DtypeType=None, order: str=None) -> ndarray[Any]: ...
def asanyarray(a: Any, dtype: DtypeType=None, order: str=None) -> ndarray[Any]: ...  # TODO figure out a way to restrict the return type
def asmatrix(data: Any, dtype: DtypeType=None) -> Any: ...  # TODO define matrix
def ascontiguousarray(a: Any, dtype: DtypeType=None) -> ndarray[Any]: ...
def copy(a: Any, order: str=None)	-> ndarray[Any]: ...
def empty(shape: ShapeType, dtype: DtypeType=float, order: str='C') -> ndarray[Any]: ...
def empty_like(a: Any, dtype: Any=None, order: str='K', subok: bool=True) -> ndarray[Any]: ...
def eye(N: int, M: int=None, k: int=0, dtype: DtypeType=float) -> ndarray[Any]: ...
def frombuffer(buffer: Any, dtype: DtypeType=float, count: int=-1,  # TODO figure out a way to restrict buffer
               offset: int=0) -> ndarray[Any]: ...
def fromfile(file: object, dtype: DtypeType=float, count: int=-1, sep: str='') -> ndarray[Any]: ...  # TODO fix file definition (There's a bug in mypy io's namespace https://github.com/python/mypy/issues/1462)
def full(shape: ShapeType, fill_value: Any, dtype: DtypeType=None,
         order: str='C') -> ndarray[Any]: ...
def full_like(a: Any, fill_value: Any, dtype: DtypeType=None, order: str='C',
              subok: bool=True) -> ndarray[Any]: ...
def fromfunction(function: Callable[..., _S], shape: ShapeType, dtype: DtypeType=float) -> ndarray[_S]: ...
def fromiter(iterable: Iterator[Any], dytpe: DtypeType, count: int=-1) -> ndarray[Any]: ...
def fromstring(string: str, dtype: DtypeType=float, count: int=-1, sep: str='') -> ndarray[Any]: ...
def identity(n: int, dtype: DtypeType=None) -> ndarray[Any]: ...
def loadtxt(fname: Any, dtype: DtypeType=float, comments: Union[str, Sequence[str]]='#',
            delimiter: str=None, converters: Dict[int, Callable[[Any], float]]=None,
            skiprows: int=0, usecols: Sequence[int]=None,
            unpack: bool=False, ndmin: int=0) -> ndarray[float]: ...
def ones(shape: ShapeType, dtype: Optional[DtypeType]=..., order: str='C') -> ndarray[Any]: ...
def ones_like(a: Any, dtype: Any=None, order: str='K', subok: bool=True) -> ndarray[Any]: ...
def zeros(shape: ShapeType, dtype: DtypeType=float, order: str='C') -> ndarray[Any]: ...
def zeros_like(a: Any, dtype: Any=None, order: str='K', subok: bool=True) -> ndarray[Any]: ...


#!/usr/bin/env python3

# pylint: disable=C0111

from collections import OrderedDict
import gzip
import os

import pytest

from pyndl.preprocess import (create_event_file, filter_event_file,
                              create_binary_event_files, bandsample,
                              event_generator, write_events,
                              _job_binary_event_file, JobFilter, to_bytes, to_integer, read_binary_file)

from pyndl.count import (cues_outcomes, load_counter, save_counter)
from pyndl import io

TEST_ROOT = os.path.join(os.path.pardir, os.path.dirname(__file__))
EVENT_FILE = os.path.join(TEST_ROOT, "temp/events_corpus.tab.gz")
RESOURCE_FILE = os.path.join(TEST_ROOT, "resources/corpus.txt")


def test_bandsample():
    resource_file = os.path.join(TEST_ROOT, "resources/event_file_trigrams_to_word.tab.gz")
    _, _, outcome_freq_map = cues_outcomes(resource_file,
                                           n_jobs=2)
    outcome_freq_map_filtered = bandsample(outcome_freq_map, 50, cutoff=1, seed=None, verbose=False)
    assert len(outcome_freq_map_filtered) == 50

    reference_file = os.path.join(TEST_ROOT, 'reference/bandsampled_outcomes.tab')
    try:
        outcome_freq_map_filtered_reference = load_counter(reference_file)
    except FileNotFoundError:
        temp_file = os.path.join(TEST_ROOT, 'temp/bandsampled_outcomes.tab')
        save_counter(outcome_freq_map_filtered, temp_file)
        raise

    bandsample(outcome_freq_map, 50, cutoff=1, verbose=True)


def test_bad_symbols():
    with pytest.raises(ValueError):
        create_event_file(RESOURCE_FILE, EVENT_FILE,
                          "abcd#")
    assert not os.path.isfile(EVENT_FILE)
    with pytest.raises(ValueError):
        create_event_file(RESOURCE_FILE, EVENT_FILE,
                          "abcd_")
    assert not os.path.isfile(EVENT_FILE)


def test_bad_event_context():
    with pytest.raises(NotImplementedError):
        create_event_file(RESOURCE_FILE, EVENT_FILE,
                          context_structure="UNREASONABLE")
    assert not os.path.isfile(EVENT_FILE)


def test_bad_event_event():
    with pytest.raises(NotImplementedError):
        create_event_file(RESOURCE_FILE, EVENT_FILE,
                          event_structure="UNREASONABLE")
    assert not os.path.isfile(EVENT_FILE)


def test_upper_case():
    event_file = os.path.join(TEST_ROOT, "temp/events_corpus_upper_case.tab.gz")
    create_event_file(RESOURCE_FILE, event_file,
                      context_structure="document",
                      event_structure="consecutive_words",
                      event_options=(3, ))
    os.remove(event_file)


def test_trigrams_to_word():
    event_file = os.path.join(TEST_ROOT, "temp/event_file_trigrams_to_word.tab.gz")
    reference_file = os.path.join(TEST_ROOT, "reference/event_file_trigrams_to_word.tab.gz")
    create_event_file(RESOURCE_FILE, event_file,
                      context_structure="document",
                      event_structure="consecutive_words",
                      event_options=(3, ),
                      cue_structure="trigrams_to_word")
    compare_event_files(event_file, reference_file)
    os.remove(event_file)


def test_trigrams_to_word_line_based():
    event_file = os.path.join(TEST_ROOT, "temp/event_file_trigrams_to_word_line_based.tab.gz")
    reference_file = os.path.join(TEST_ROOT, "reference/event_file_trigrams_to_word_line_based.tab.gz")
    create_event_file(RESOURCE_FILE, event_file,
                      context_structure="document",
                      event_structure="line", event_options=(3, ),
                      cue_structure="trigrams_to_word")
    compare_event_files(event_file, reference_file)
    os.remove(event_file)


def test_bigrams_to_word():
    event_file = os.path.join(TEST_ROOT, "temp/event_file_bigrams_to_word.tab.gz")
    reference_file = os.path.join(TEST_ROOT, "reference/event_file_bigrams_to_word.tab.gz")
    create_event_file(RESOURCE_FILE, event_file,
                      context_structure="document",
                      event_structure="consecutive_words",
                      event_options=(3, ),
                      cue_structure="bigrams_to_word",
                      remove_duplicates=True)
    compare_event_files(event_file, reference_file)
    os.remove(event_file)


def test_remove_duplicates():
    event_file_noduplicates = os.path.join(TEST_ROOT, "temp/event_file_bigrams_to_word_noduplicates.tab.gz")
    event_file_duplicates = os.path.join(TEST_ROOT, "temp/event_file_bigrams_to_word_duplicates.tab.gz")
    create_event_file(RESOURCE_FILE, event_file_duplicates,
                      context_structure="document",
                      event_structure="consecutive_words",
                      event_options=(3, ),
                      cue_structure="bigrams_to_word",
                      remove_duplicates=False)
    create_event_file(RESOURCE_FILE, event_file_noduplicates,
                      context_structure="document",
                      event_structure="consecutive_words",
                      event_options=(3, ),
                      cue_structure="bigrams_to_word",
                      remove_duplicates=True)

    with gzip.open(event_file_noduplicates, "rt") as new_file:
        lines_new = new_file.readlines()
    with gzip.open(event_file_duplicates, "rt") as reference:
        lines_reference = reference.readlines()
    assert len(lines_new) == len(lines_reference)
    n_cues_unequal = 0
    n_outcomes_unequal = 0
    for ii, line in enumerate(lines_new):
        cues, outcomes = line.strip().split('\t')
        cues = sorted(cues.split('_'))
        outcomes = sorted(outcomes.split('_'))
        ref_cues, ref_outcomes = lines_reference[ii].strip().split('\t')
        ref_cues = sorted(ref_cues.split('_'))
        ref_outcomes = sorted(ref_outcomes.split('_'))
        if len(cues) != len(ref_cues):
            n_cues_unequal += 1
        if len(outcomes) != len(ref_outcomes):
            n_outcomes_unequal += 1
        # there should be no duplicates in (noduplicates)
        assert len(cues) == len(set(cues))
        assert len(outcomes) == len(set(outcomes))
        # after making each list unique it should be the same
        assert set(cues) == set(ref_cues)
        assert set(outcomes) == set(ref_outcomes)
    assert n_cues_unequal == 1098
    assert n_outcomes_unequal == 66

    os.remove(event_file_noduplicates)
    os.remove(event_file_duplicates)


def test_word_to_word():
    event_file = os.path.join(TEST_ROOT, "temp/event_file_word_to_word.tab.gz")
    reference_file = os.path.join(TEST_ROOT, "reference/event_file_word_to_word.tab.gz")
    create_event_file(RESOURCE_FILE, event_file,
                      context_structure="document",
                      event_structure="word_to_word", event_options=(2, 1),
                      cue_structure="word_to_word")
    compare_event_files(event_file, reference_file)

    os.remove(event_file)


def test_bad_event_file():
    input_event_file = os.path.join(TEST_ROOT, "resources/event_file_trigrams_to_word_BAD.tab.gz")
    output_event_file = os.path.join(TEST_ROOT, "temp/event_file_BAD_output.tab.gz")
    with pytest.raises(ValueError):
        filter_event_file(input_event_file, output_event_file)
    os.remove(output_event_file)


def test_job_filter():
    keep_cues = ["#of", "of#"]
    keep_outcomes = ["of", ]
    job = JobFilter(keep_cues, keep_outcomes, None, None, None, None)
    line = '#of_alb_NEI_b_of#_XX\tterm_not_of\n'
    new_line = job.job(line)
    assert new_line == '#of_of#\tof\n'
    # no cues
    line = 'alb_NEI_b_XX\tterm_not_of\n'
    new_line = job.job(line)
    assert new_line is None
    # no outcomes
    line = '#of_alb_NEI_b_of#_XX\tterm_not\n'
    new_line = job.job(line)
    assert new_line == '#of_of#\t\n'
    # neither cues nor outcomes
    line = '#alb_NEI_b_XX\tterm_not\n'
    new_line = job.job(line)
    assert new_line is None
    with pytest.raises(ValueError):
        bad_line = 'This is a bad line.'
        job.job(bad_line)


def test_filter_event_file():
    input_event_file = os.path.join(TEST_ROOT, "resources/event_file_trigrams_to_word.tab.gz")
    output_event_file = os.path.join(TEST_ROOT, "temp/event_file_filtered.tab.gz")
    cues = ["#of", "of#"]
    cues.sort()
    outcomes = ["of", ]
    outcomes.sort()
    filter_event_file(input_event_file, output_event_file,
                      keep_cues=cues,
                      keep_outcomes=outcomes,
                      n_jobs=2,
                      verbose=True)
    _, cue_freq_map, outcome_freq_map = cues_outcomes(output_event_file)
    cues_new = list(cue_freq_map)
    cues_new.sort()
    outcomes_new = list(outcome_freq_map)
    outcomes_new.sort()
    assert cues == cues_new
    assert outcomes == outcomes_new
    os.remove(output_event_file)


def test_write_events():
    event_file = os.path.join(TEST_ROOT, "resources/event_file_trigrams_to_word.tab.gz")
    n_events, cue_freq_map, outcome_freq_map = cues_outcomes(event_file)
    outcomes = list(outcome_freq_map.keys())
    outcomes.sort()
    cues = list(cue_freq_map.keys())
    cues.sort()
    cue_id_map = {cue: ii for ii, cue in enumerate(cues)}
    outcome_id_map = {outcome: nn for nn, outcome in enumerate(outcomes)}
    events = event_generator(event_file, cue_id_map, outcome_id_map, sort_within_event=True)
    file_name = os.path.join(TEST_ROOT, "temp/events.bin")
    with pytest.raises(StopIteration):
        write_events(events, file_name, remove_duplicates=True)
    os.remove(file_name)

    # start stop
    events = event_generator(event_file, cue_id_map, outcome_id_map, sort_within_event=True)
    n_events = write_events(events, file_name, start=10, stop=20, remove_duplicates=True)
    assert n_events == 10
    os.remove(file_name)

    # no events
    events = event_generator(event_file, cue_id_map, outcome_id_map, sort_within_event=True)
    n_events = write_events(events, file_name, start=100000, stop=100010, remove_duplicates=True)
    assert n_events == 0

    _job_binary_event_file(file_name=file_name, event_file=event_file,
                           cue_id_map=cue_id_map,
                           outcome_id_map=outcome_id_map,
                           sort_within_event=False,
                           start=0, stop=10, remove_duplicates=True)
    _job_binary_event_file(file_name=file_name, event_file=event_file,
                           cue_id_map=cue_id_map,
                           outcome_id_map=outcome_id_map,
                           sort_within_event=False,
                           start=0, stop=10, remove_duplicates=True)
    os.remove(file_name)

    # bad event file
    with pytest.raises(ValueError):
        event_bad_file = os.path.join(TEST_ROOT, "resources/event_file_trigrams_to_word_BAD.tab.gz")
        events = event_generator(event_bad_file, cue_id_map,
                                 outcome_id_map)
        # traverse generator
        # pylint: disable=W0612
        for event in events:
            pass


def test_byte_conversion():
    simple_int = 184729172
    assert simple_int == to_integer(to_bytes(simple_int))


def test_read_binary_file():
    file_path = "resources/event_file_trigrams_to_word.tab.gz"
    binary_path = "binary_resources/"

    abs_file_path = os.path.join(TEST_ROOT, file_path)
    abs_binary_path = os.path.join(TEST_ROOT, binary_path)
    abs_binary_file_path = os.path.join(abs_binary_path, "events_0_0.dat")

    _, cues, outcomes = cues_outcomes(abs_file_path)
    cue_id_map = OrderedDict(((cue, ii) for ii, cue in enumerate(cues.keys())))
    outcome_id_map = OrderedDict(((outcome, ii) for ii, outcome in enumerate(outcomes.keys())))

    number_events = create_binary_event_files(abs_file_path, abs_binary_path, cue_id_map,
                                              outcome_id_map, overwrite=True, remove_duplicates=False)

    bin_events = read_binary_file(abs_binary_file_path)
    events = io.events_from_file(abs_file_path)
    events_dup = io.events_from_file(abs_file_path)

    assert number_events == len(list(events_dup))

    for event, bin_event in zip(events, bin_events):
        cues, outcomes = event
        bin_cues, bin_outcomes = bin_event
        if len(cues) != len(bin_cues):
            raise ValueError('Cues have different length')
        if len(outcomes) != len(bin_outcomes):
            raise ValueError('Cues have different length')

        for cue, bin_cue in zip(cues, bin_cues):
            assert cue_id_map[cue] == bin_cue

        for outcome, bin_outcome in zip(outcomes, bin_outcomes):
            assert outcome_id_map[outcome] == bin_outcome

    # clean everything
    os.remove(abs_binary_file_path)


def test_preprocessing():
    corpus_file = os.path.join(TEST_ROOT, "resources/corpus.txt")
    event_file = os.path.join(TEST_ROOT, "temp/events_corpus.tab.gz")
    symbols = "abcdefghijklmnopqrstuvwxyzóąćęłńśźż"  # polish

    # create event file
    create_event_file(corpus_file, event_file, symbols,
                      context_structure="document",
                      event_structure="consecutive_words",
                      event_options=(3, ),
                      lower_case=True, verbose=True)

    # read in outcomes
    _, _, outcome_freq_map = cues_outcomes(event_file, n_jobs=2)

    # reduce number of outcomes through bandsampling
    outcome_freq_map_filtered = bandsample(outcome_freq_map, 50, cutoff=1, seed=None)
    outcomes = list(outcome_freq_map_filtered.keys())
    outcomes.sort()

    # filter outcomes by reduced number of outcomes
    event_file_filtered = event_file + ".filtered"
    filter_event_file(event_file, event_file_filtered, keep_outcomes=outcomes)

    # clean everything
    os.remove(event_file)
    os.remove(event_file_filtered)


def compare_event_files(newfile, oldfile):
    with gzip.open(newfile, "rt") as new_file:
        lines_new = new_file.readlines()
    with gzip.open(oldfile, "rt") as reference:
        lines_reference = reference.readlines()
    assert len(lines_new) == len(lines_reference)
    for ii, line in enumerate(lines_new):
        cues, outcomes = line.strip().split('\t')
        cues = sorted(cues.split('_'))
        outcomes = sorted(outcomes.split('_'))
        ref_cues, ref_outcomes = lines_reference[ii].strip().split('\t')
        ref_cues = sorted(ref_cues.split('_'))
        ref_outcomes = sorted(ref_outcomes.split('_'))
        assert cues == ref_cues
        assert outcomes == ref_outcomes



"""
pyndl.activation
----------------

*pyndl.activation* provides the functionality to estimate activation of a
trained ndl model for given events. The trained ndl model is thereby
represented as the outcome-cue weights.
"""
import multiprocessing as mp
import ctypes
from collections import defaultdict, OrderedDict
import warnings

import numpy as np
import xarray as xr

from . import io


# pylint: disable=W0621
def activation(events, weights, *, n_jobs=1, number_of_threads=None,
               remove_duplicates=None, ignore_missing_cues=False):
    """
    Estimate activations for given events in event file and outcome-cue weights.

    Memory overhead for multiprocessing is one copy of weights
    plus a copy of cues for each thread.

    Parameters
    ----------
    events : generator or str
        generates cues, outcomes pairs or the path to the event file
    weights : xarray.DataArray or dict[dict[float]]
        the xarray.DataArray needs to have the dimensions 'outcomes' and 'cues'
        the dictionaries hold weight[outcome][cue].
    n_jobs : int
        a integer giving the number of threads in which the job should
        executed
    remove_duplicates : {None, True, False}
        if None raise a ValueError when the same cue is present multiple times
        in the same event; True make cues unique per event; False
        keep multiple instances of the same cue (this is usually not
        preferred!)
    ignore_missing_cues : {True, False}
        if True function ignores cues which are in the test dataset but not in
        the weight matrix
        if False raises a KeyError for cues which are not in the weight matrix

    Returns
    -------
    activations : xarray.DataArray
        with dimensions 'outcomes' and 'events'. Contains coords for the outcomes.
        returned if weights is instance of xarray.DataArray

    or

    activations : dict of numpy.arrays
        the first dict has outcomes as keys and dicts as values
        the list has a activation value per event
        returned if weights is instance of dict

    """
    if number_of_threads is not None:
        warnings.warn("Parameter `number_of_threads` is renamed to `n_jobs`. The old name "
                      "will stop working with v0.9.0.",
                      DeprecationWarning, stacklevel=2)
        n_jobs = number_of_threads
    if isinstance(events, str):
        events = io.events_from_file(events)

    events = (cues for cues, outcomes in events)
    if remove_duplicates is None:
        def check_no_duplicates(cues):
            if len(cues) != len(set(cues)):
                raise ValueError('cues needs to be unique: "{}"; use '
                                 'remove_duplicates=True'.format(' '.join(cues)))
            else:
                return set(cues)
        events = (check_no_duplicates(cues) for cues in events)
    elif remove_duplicates is True:
        events = (set(cues) for cues in events)

    if isinstance(weights, xr.DataArray):
        cues = weights.coords["cues"].values.tolist()
        outcomes = weights.coords["outcomes"].values.tolist()
        if not weights.values.shape == (len(outcomes), len(cues)):
            raise ValueError('dimensions of weights are wrong. Probably you need to transpose the matrix')
        cue_map = OrderedDict(((cue, ii) for ii, cue in enumerate(cues)))
        if ignore_missing_cues:
            event_cue_indices_list = (tuple(cue_map[cue] for cue in event_cues if cue in cues)
                                      for event_cues in events)
        else:
            event_cue_indices_list = (tuple(cue_map[cue] for cue in event_cues)
                                      for event_cues in events)
        # pylint: disable=W0621
        activations = _activation_matrix(list(event_cue_indices_list),
                                         weights.values, n_jobs)
        return xr.DataArray(activations,
                            coords={
                                'outcomes': outcomes
                            },
                            dims=('outcomes', 'events'))
    elif isinstance(weights, dict):
        assert n_jobs == 1, "Estimating activations with multiprocessing is not implemented for dicts."
        activations = defaultdict(lambda: np.zeros(len(events)))
        events = list(events)
        for outcome, cue_dict in weights.items():
            _activations = activations[outcome]
            for row, cues in enumerate(events):
                for cue in cues:
                    _activations[row] += cue_dict[cue]
        return activations
    else:
        raise ValueError("Weights other than xarray.DataArray or dicts are not supported.")


def _init_mp_activation_matrix(weights_, weights_shape_, activations_, activations_shape_):
    """
    Private helper function for multiprocessing in _activation_matrix.
    Initializes shared variables weights and activations.

    """
    # pylint: disable=C0103, W0621, W0601
    global weights, activations
    weights = np.ctypeslib.as_array(weights_)
    weights.shape = weights_shape_
    activations = np.ctypeslib.as_array(activations_)
    activations.shape = activations_shape_


def _run_mp_activation_matrix(event_index, cue_indices):
    """
    Private helper function for multiprocessing in _activation_matrix.
    Calculate activation for all outcomes while a event.

    """
    activations[:, event_index] = weights[:, cue_indices].sum(axis=1)


def _activation_matrix(indices_list, weights, n_jobs):
    """
    Estimate activation for indices in weights

    Memory overhead for multiprocessing is one copy of weights
    plus a copy of cues for each thread.

    Parameters
    ----------
    indices_list : list[int]
        events as cue indices in weights
    weights : numpy.array
        weight matrix with shape (outcomes, cues)
    n_jobs : int

    Returns
    -------
    activation_matrix : numpy.array
        estimated activations as matrix with shape (outcomes, events)

    """
    assert n_jobs >= 1, "Can't run with less than 1 thread"

    activations_dim = (weights.shape[0], len(indices_list))
    if n_jobs == 1:
        activations = np.empty(activations_dim, dtype=np.float64)
        for row, event_cues in enumerate(indices_list):
            activations[:, row] = weights[:, event_cues].sum(axis=1)
        return activations
    else:
        shared_activations = mp.RawArray(ctypes.c_double, int(np.prod(activations_dim)))
        weights = np.ascontiguousarray(weights)
        shared_weights = mp.sharedctypes.copy(np.ctypeslib.as_ctypes(np.float64(weights)))
        initargs = (shared_weights, weights.shape, shared_activations, activations_dim)
        with mp.Pool(n_jobs, initializer=_init_mp_activation_matrix, initargs=initargs) as pool:
            pool.starmap(_run_mp_activation_matrix, enumerate(indices_list))
        activations = np.ctypeslib.as_array(shared_activations)
        activations.shape = activations_dim
        return activations



cdef enum ErrorCode:
    NO_ERROR = 0
    MAGIC_NUMBER_DOES_NOT_MATCH = 1
    VERSION_NUMBER_DOES_NOT_MATCH = 2
    INITIAL_ERROR_CODE = 3
    ONLY_ONE_OUTCOME_PER_EVENT = 7


ERROR_CODES = """
    NO_ERROR = 0
    MAGIC_NUMBER_DOES_NOT_MATCH = 1
    VERSION_NUMBER_DOES_NOT_MATCH = 2
    INITIAL_ERROR_CODE = 3
    ONLY_ONE_OUTCOME_PER_EVENT = 7
    """



import sys
import time

import numpy as np

from scipy import stats

if sys.platform.startswith('linux'):
    from . import correlation_openmp
elif sys.platform.startswith('win32'):
    pass
elif sys.platform.startswith('darwin'):
    pass


def _reference_correlation(semantics, activations, *, verbose=False):
    """
    calculates the correlations between the semantics and the activations.

    Returns
    -------
    np.array (n_outcomes, n_events)

    The first column contains all correlations between the first event and
    all possible outcomes in the semantcs.

    The first column reads like:

    0. correlation between first event and first outcome in the semantic
       (gold standard) space.
    1. correlation between first event and second outcome ...
    ...

    """
    assert semantics.shape[0] == activations.shape[0], ("number of vector dimensions in semantics and activations"
                                                        " need to be the same")
    n_outcomes = semantics.shape[1]
    n_events = activations.shape[1]

    correlations = np.zeros((n_outcomes, n_events))

    start_time = time.time()
    for ii in range(n_events):
        for jj in range(n_outcomes):
            correlations[jj, ii], _ = stats.pearsonr(semantics[:, jj], activations[:, ii])
    if verbose:
        print(f"time needed for correlations:  {time.time() - start_time}")

    return correlations


def correlation(semantics, activations, *, verbose=False, allow_nan=False):
    """
    calculates the correlations between the semantics and the activations.

    Returns
    -------
    np.array (n_outcomes, n_events)

    The first column contains all correlations between the first event and
    all possible outcomes in the semantcs.

    The first column reads like:

    0. correlation between first event and first outcome in the semantic
       (gold standard) space.
    1. correlation between first event and second outcome ...
    ...

    """
    if not sys.platform.startswith('linux'):
        raise NotImplementedError("OpenMP is linux only at the moment.")

    assert semantics.shape[0] == activations.shape[0], ("number of vector dimensions in semantics and activations"
                                                        "need to be the same")
    n_outcomes = semantics.shape[1]
    n_vec_dims, n_events = activations.shape

    semantics_means = np.zeros((n_outcomes,))
    semantics_stds = np.zeros((n_outcomes,))
    activations_means = np.zeros((n_events,))
    activations_stds = np.zeros((n_events,))

    if verbose:
        start_time = time.time()

    for jj in range(n_outcomes):
        semantics_means[jj] = np.mean(semantics[:, jj])
        semantics_stds[jj] = np.std(semantics[:, jj], ddof=1)

    for ii in range(n_events):
        activations_means[ii] = np.mean(activations[:, ii])
        activations_stds[ii] = np.std(activations[:, ii], ddof=1)

    if verbose:
        print(f"time needed for stds and means:  {time.time() - start_time}")

    if not allow_nan:
        if np.any(semantics_stds == 0) or np.any(np.isnan(semantics_stds)):
            raise ValueError('Standard deviations of semantics are not different to zero or nan.')
        if np.any(activations_stds == 0) or np.any(np.isnan(activations_stds)):
            raise ValueError('Standard deviations of activations are not different to zero or nan.')

    if verbose:
        start_time = time.time()

    correlations = correlation_openmp.correlation(semantics, activations, semantics_means,
                                                  semantics_stds, activations_means, activations_stds)
    if verbose:
        print(f"time needed for correlations:  {time.time() - start_time}")

    return correlations



#!/usr/bin/env python3

# pylint: disable=C0111, redefined-outer-name


from collections import defaultdict, OrderedDict
import os
import tempfile

import numpy as np
import xarray as xr
import pandas as pd
import pytest

from pyndl import ndl, count, wh

TEST_ROOT = os.path.join(os.path.pardir, os.path.dirname(__file__))
FILE_PATH_WH = os.path.join(TEST_ROOT, "resources/event_file_wh.tab.gz")

TMP_PATH = tempfile.mkdtemp()

ETA = 0.01

CONTINUE_SPLIT_POINT = 3


@pytest.mark.nolinux
def test_consistency_wh():
    events = FILE_PATH_WH

    cue_vectors = xr.DataArray(np.array([[0.2, 0.11, 0.5, 0, 0], [0, 0, 0.5, 0.11, 0.2], [0.2, 0, 0, 0, 0.2]]),
                               dims=('cues', 'cue_vector_dimensions'),
                               coords={'cues': ['a', 'b', 'c'],
                                       'cue_vector_dimensions': ['dim1', 'dim2', 'dim3', 'dim4', 'dim5']})
    outcome_vectors = xr.DataArray(np.array([[0.2, 1.], [0.5, 0], [0, 0.5], [1., 0.2]]),
                                   dims=('outcomes', 'outcome_vector_dimensions'),
                                   coords={'outcomes': ['A', 'B', 'C', 'D'],
                                           'outcome_vector_dimensions': ['o_dim1', 'o_dim2']})

    _ = wh.dict_wh(events, ETA, cue_vectors=cue_vectors, outcome_vectors=outcome_vectors)
    weights = wh.dict_wh(events, ETA, cue_vectors=cue_vectors, outcome_vectors=outcome_vectors,
                         make_data_array=True, verbose=True)

    weights_np = wh.wh(events, ETA, cue_vectors=cue_vectors, outcome_vectors=outcome_vectors,
                       method='numpy', verbose=True)
    weights_openmp = wh.wh(events, ETA, cue_vectors=cue_vectors,
                           outcome_vectors=outcome_vectors, method='openmp', verbose=True)
    assert np.all(weights == weights_np)
    assert np.all(weights == weights_openmp)

    cue_vectors = xr.DataArray(np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]]),
                               dims=('cues', 'cue_vector_dimensions'),
                               coords={'cues': ['a', 'b', 'c'], 'cue_vector_dimensions': ['dim1', 'dim2', 'dim3']})
    outcome_vectors = xr.DataArray(np.array([[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 1, 0], [0, 0, 0, 1]]),
                                   dims=('outcomes', 'outcome_vector_dimensions'),
                                   coords={'outcomes': ['A', 'B', 'C', 'D'],
                                           'outcome_vector_dimensions': ['dim1', 'dim2', 'dim3', 'dim4']})

    weights_wh = wh.dict_wh(events, ETA, cue_vectors, outcome_vectors, make_data_array=True)
    weights_ndl = ndl.dict_ndl(events, alphas=defaultdict(lambda: 1), betas=(ETA, ETA), make_data_array=True)

    weights_wh = weights_wh.loc[{'outcome_vector_dimensions': ['dim1', 'dim2', 'dim3', 'dim4'],
                                 'cue_vector_dimensions': ['dim1', 'dim2', 'dim3']}]
    weights_wh.coords['outcome_vector_dimensions'] = ['A', 'B', 'C', 'D']
    weights_wh.coords['cue_vector_dimensions'] = ['a', 'b', 'c']
    weights_wh = weights_wh.rename({'outcome_vector_dimensions': 'outcomes', 'cue_vector_dimensions': 'cues'})
    unequal, unequal_ratio = compare_arrays(events, weights_wh, weights_ndl)
    print(unequal)
    print('%.2f ratio unequal' % unequal_ratio)
    assert len(unequal) == 0  # pylint: disable=len-as-condition


@pytest.mark.nolinux
@pytest.mark.runslow
def test_real_to_real_wh_large():
    events = FILE_PATH_WH
    eta = 0.001

    n_cue_vec_dims = 1000
    n_outcome_vec_dims = 4000
    cue_vectors = xr.DataArray(np.random.random((3, n_cue_vec_dims)), dims=('cues', 'cue_vector_dimensions'),
                               coords={'cues': ['a', 'b', 'c'],
                                       'cue_vector_dimensions': [f'c_dim{ii}' for ii in range(n_cue_vec_dims)]})
    outcome_vectors = xr.DataArray(np.random.random((4, n_outcome_vec_dims)),
                                   dims=('outcomes', 'outcome_vector_dimensions'),
                                   coords={'outcomes': ['A', 'B', 'C', 'D'],
                                           'outcome_vector_dimensions': [f'o_dim{ii}' for ii
                                                                         in range(n_outcome_vec_dims)]})

    # 1 min 41 sec
    weights_np = wh.wh(events, eta, cue_vectors=cue_vectors,
                       outcome_vectors=outcome_vectors, method='numpy')
    # 18 sec
    weights_openmp = wh.wh(events, eta, cue_vectors=cue_vectors,
                           outcome_vectors=outcome_vectors, method='openmp')
    assert np.allclose(weights_openmp.data, weights_np.data)

    # 1 min 41 sec
    weights_split_small = wh.wh(events, eta, outcome_vectors=outcome_vectors,
                                n_outcomes_per_job=100)
    # 18 sec
    weights_openmp = wh.wh(events, eta, outcome_vectors=outcome_vectors, method='openmp')
    assert np.allclose(weights_openmp.data, weights_split_small.data)


@pytest.mark.nolinux
@pytest.mark.runslow
def test_binary_to_real_wh_continue_learning():
    cue_vectors = xr.DataArray(np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]], dtype=float),
                               dims=('cues', 'cue_vector_dimensions'),
                               coords={'cues': ['a', 'b', 'c'],
                                       'cue_vector_dimensions': ['c_dim1', 'c_dim2', 'c_dim3']})
    outcome_vectors = xr.DataArray(np.array([[0.2, 1., 0.7, -0.2], [0.5, 0, 0.1, 0.0],
                                             [0, 0.5, 0.0, 0.1], [1., 0.2, -0.8, 0.11]]),
                                   dims=('outcomes', 'outcome_vector_dimensions'),
                                   coords={'outcomes': ['A', 'B', 'C', 'D'],
                                           'outcome_vector_dimensions': ['o_dim1', 'o_dim2', 'o_dim3', 'o_dim4']})

    weights = wh.wh(FILE_PATH_WH, ETA, outcome_vectors=outcome_vectors, verbose=True)
    for ii in range(10):
        weights = wh.wh(FILE_PATH_WH, ETA, outcome_vectors=outcome_vectors,
                        weights=weights, n_jobs=ii % 4 + 1,
                        n_outcomes_per_job=4 - ii % 4)

    weights_real_to_real = wh.wh(FILE_PATH_WH, ETA, cue_vectors=cue_vectors,
                                 outcome_vectors=outcome_vectors)
    for _ in range(10):
        weights_real_to_real = wh.wh(FILE_PATH_WH, ETA,
                                     cue_vectors=cue_vectors, outcome_vectors=outcome_vectors,
                                     weights=weights_real_to_real)

    weights_real_to_real.coords['cue_vector_dimensions'] = ['a', 'b', 'c']
    weights_real_to_real = weights_real_to_real.rename({'cue_vector_dimensions': 'cues'})

    unequal, unequal_ratio = compare_arrays(FILE_PATH_WH, weights, weights_real_to_real)
    print(unequal)
    print('%.2f ratio unequal' % unequal_ratio)
    assert len(unequal) == 0  # pylint: disable=len-as-condition


@pytest.mark.nolinux
def test_binary_to_real_wh():
    cue_vectors = xr.DataArray(np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]], dtype=float),
                               dims=('cues', 'cue_vector_dimensions'),
                               coords={'cues': ['a', 'b', 'c'],
                                       'cue_vector_dimensions': ['c_dim1', 'c_dim2', 'c_dim3']})
    outcome_vectors = xr.DataArray(np.array([[0.2, 1., 0.7, -0.2], [0.5, 0, 0.1, 0.0],
                                             [0, 0.5, 0.0, 0.1], [1., 0.2, -0.8, 0.11]]),
                                   dims=('outcomes', 'outcome_vector_dimensions'),
                                   coords={'outcomes': ['A', 'B', 'C', 'D'],
                                           'outcome_vector_dimensions': ['o_dim1', 'o_dim2', 'o_dim3', 'o_dim4']})

    weights = wh.wh(FILE_PATH_WH, ETA, outcome_vectors=outcome_vectors, verbose=True)
    # weights.to_netcdf(os.path.join(TEST_ROOT, 'reference/binary_to_real_weights.nc'))
    reference_weights = xr.open_dataarray(os.path.join(TEST_ROOT, 'reference/binary_to_real_weights.nc'))

    unequal, unequal_ratio = compare_arrays(FILE_PATH_WH, weights, reference_weights)
    print(unequal)
    print('%.2f ratio unequal' % unequal_ratio)
    assert len(unequal) == 0  # pylint: disable=len-as-condition

    weights_real_to_real = wh.wh(FILE_PATH_WH, ETA, cue_vectors=cue_vectors, outcome_vectors=outcome_vectors)

    weights_real_to_real.coords['cue_vector_dimensions'] = ['a', 'b', 'c']
    weights_real_to_real = weights_real_to_real.rename({'cue_vector_dimensions': 'cues'})

    unequal, unequal_ratio = compare_arrays(FILE_PATH_WH, weights, weights_real_to_real)
    print(unequal)
    print('%.2f ratio unequal' % unequal_ratio)
    assert len(unequal) == 0  # pylint: disable=len-as-condition

    weights_split = wh.wh(FILE_PATH_WH, ETA, outcome_vectors=outcome_vectors, verbose=True, n_outcomes_per_job=2)
    unequal, unequal_ratio = compare_arrays(FILE_PATH_WH, weights, weights_split)
    print(unequal)
    print('%.2f ratio unequal' % unequal_ratio)
    assert len(unequal) == 0  # pylint: disable=len-as-condition

    for _ in range(1):
        weights_split = wh.wh(FILE_PATH_WH, ETA, weights=weights_split,
                              outcome_vectors=outcome_vectors, verbose=True,
                              n_outcomes_per_job=2)
        weights = wh.wh(FILE_PATH_WH, ETA, weights=weights,
                        outcome_vectors=outcome_vectors, verbose=True)
    unequal, unequal_ratio = compare_arrays(FILE_PATH_WH, weights, weights_split)
    print(unequal)
    print('%.2f ratio unequal' % unequal_ratio)
    print(weights_split)
    print(weights)
    assert len(unequal) == 0  # pylint: disable=len-as-condition


@pytest.mark.nolinux
def test_real_to_binary_wh():
    cue_vectors = xr.DataArray(np.array([[0.2, 0.11, 0.5, 0, 0], [0, 0, 0.5, 0.11, 0.2], [0.2, 0, 0, 0, 0.2]]),
                               dims=('cues', 'cue_vector_dimensions'),
                               coords={'cues': ['a', 'b', 'c'],
                                       'cue_vector_dimensions': ['c_dim1', 'c_dim2', 'c_dim3', 'c_dim4', 'c_dim5']})
    outcome_vectors = xr.DataArray(np.array([[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 1, 0], [0, 0, 0, 1]], dtype=float),
                                   dims=('outcomes', 'outcome_vector_dimensions'),
                                   coords={'outcomes': ['A', 'B', 'C', 'D'],
                                           'outcome_vector_dimensions': ['dim1', 'dim2', 'dim3', 'dim4']})

    weights = wh.wh(FILE_PATH_WH, ETA, cue_vectors=cue_vectors, verbose=True)
    # weights.to_netcdf(os.path.join(TEST_ROOT, 'reference/real_to_binary_weights.nc'))
    reference_weights = xr.open_dataarray(os.path.join(TEST_ROOT, 'reference/real_to_binary_weights.nc'))

    unequal, unequal_ratio = compare_arrays(FILE_PATH_WH, weights, reference_weights)
    print(unequal)
    print('%.2f ratio unequal' % unequal_ratio)
    assert len(unequal) == 0  # pylint: disable=len-as-condition

    weights_real_to_real = wh.wh(FILE_PATH_WH, ETA, cue_vectors=cue_vectors, outcome_vectors=outcome_vectors)

    weights_real_to_real.coords['outcome_vector_dimensions'] = ['A', 'B', 'C', 'D']
    weights_real_to_real = weights_real_to_real.rename({'outcome_vector_dimensions': 'outcomes'})

    unequal, unequal_ratio = compare_arrays(FILE_PATH_WH, weights, weights_real_to_real)
    print(unequal)
    print('%.2f ratio unequal' % unequal_ratio)
    assert len(unequal) == 0  # pylint: disable=len-as-condition


@pytest.mark.nolinux
def test_continue_learning_wh():
    events = FILE_PATH_WH

    cue_vectors = xr.DataArray(np.array([[0.2, 0.11, 0.5, 0, 0], [0, 0, 0.5, 0.11, 0.2], [0.2, 0, 0, 0, 0.2]]),
                               dims=('cues', 'cue_vector_dimensions'),
                               coords={'cues': ['a', 'b', 'c'],
                                       'cue_vector_dimensions': ['dim1', 'dim2', 'dim3', 'dim4', 'dim5']})
    outcome_vectors = xr.DataArray(np.array([[0.2, 1.], [0.5, 0], [0, 0.5], [1., 0.2]]),
                                   dims=('outcomes', 'outcome_vector_dimensions'),
                                   coords={'outcomes': ['A', 'B', 'C', 'D'],
                                           'outcome_vector_dimensions': ['o_dim1', 'o_dim2']})

    weights = wh.wh(events, ETA, cue_vectors=cue_vectors, outcome_vectors=outcome_vectors)
    weights = wh.wh(events, ETA, cue_vectors=cue_vectors, outcome_vectors=outcome_vectors, weights=weights)
    # TODO: insert assert

    weights = wh.wh(events, ETA, outcome_vectors=outcome_vectors)
    weights = wh.wh(events, ETA, outcome_vectors=outcome_vectors, weights=weights)
    # TODO: insert assert

    weights = wh.wh(events, ETA, cue_vectors=cue_vectors)
    weights = wh.wh(events, ETA, cue_vectors=cue_vectors, weights=weights)
    # TODO: insert assert


def compare_arrays(file_path, arr1, arr2):
    _, cues, outcomes = count.cues_outcomes(file_path)
    cue_map, outcome_map, _ = generate_mapping(file_path)

    unequal = list()

    if isinstance(arr1, xr.DataArray):
        outcome_dim_name, cue_dim_name = arr1.dims

    if outcome_dim_name == 'outcome_vector_dimensions':
        outcomes = list(arr1[outcome_dim_name].data)
    if cue_dim_name == 'cue_vector_dimensions':
        cues = list(arr1[cue_dim_name].data)

    for outcome in outcomes:
        for cue in cues:
            values = list()
            for array in (arr1, arr2):
                if isinstance(array, np.ndarray):
                    outcome_index = outcome_map[outcome]
                    cue_index = cue_map[cue]
                    values.append(array[outcome_index][cue_index])
                elif isinstance(array, xr.DataArray):
                    values.append(array.loc[{outcome_dim_name: outcome, cue_dim_name: cue}].values)
                elif isinstance(array, pd.DataFrame):
                    values.append(array.loc[outcome][cue])
                else:
                    values.append(array[outcome][cue])

            value1, value2 = values  # pylint: disable=unbalanced-tuple-unpacking
            if not np.isclose(value1, value2, rtol=1e-02, atol=1e-05):
                unequal.append((outcome, cue, value1, value2))

    unequal_ratio = len(unequal) / (len(outcomes) * len(cues))
    return (unequal, unequal_ratio)


def generate_mapping(event_path):
    _, cues, outcomes = count.cues_outcomes(event_path)
    all_cues = list(cues.keys())
    all_outcomes = list(outcomes.keys())
    cue_map = OrderedDict(((cue, ii) for ii, cue in enumerate(all_cues)))
    outcome_map = OrderedDict(((outcome, ii) for ii, outcome in enumerate(all_outcomes)))

    return (cue_map, outcome_map, all_outcomes)



#!/usr/bin/env python3

# pylint: disable=C0111, redefined-outer-name


from collections import defaultdict, OrderedDict
import os
import time
import tempfile
import copy

import numpy as np
import xarray as xr
import pandas as pd
import pytest

from pyndl import ndl, count, io

TEST_ROOT = os.path.join(os.path.pardir, os.path.dirname(__file__))
FILE_PATH_SIMPLE = os.path.join(TEST_ROOT, "resources/event_file_simple.tab.gz")
FILE_PATH_MULTIPLE_CUES = os.path.join(TEST_ROOT, "resources/event_file_multiple_cues.tab.gz")
REFERENCE_PATH = os.path.join(TEST_ROOT, 'reference/weights_event_file_simple.csv')
REFERENCE_PATH_NDL2 = os.path.join(TEST_ROOT, 'reference/weights_event_file_simple_ndl2.csv')
REFERENCE_PATH_MULTIPLE_CUES_NDL2 = os.path.join(TEST_ROOT, 'reference/weights_event_file_multiple_cues_ndl2.csv')

TMP_PATH = tempfile.mkdtemp()

LAMBDA_ = 1.0
ALPHA = 0.1
BETAS = (0.1, 0.1)

CONTINUE_SPLIT_POINT = 3


@pytest.fixture(scope='module')
def result_ndl_threading():
    return ndl.ndl(FILE_PATH_SIMPLE, ALPHA, BETAS, method='threading')


@pytest.fixture(scope='module')
def result_ndl_openmp():
    return ndl.ndl(FILE_PATH_SIMPLE, ALPHA, BETAS, method='openmp')


@pytest.fixture(scope='module')
def result_dict_ndl():
    return ndl.dict_ndl(FILE_PATH_SIMPLE, ALPHA, BETAS)


@pytest.fixture(scope='module')
def result_dict_ndl_generator():
    return ndl.dict_ndl(io.events_from_file(FILE_PATH_SIMPLE), ALPHA, BETAS)


@pytest.fixture(scope='module')
def result_dict_ndl_data_array():
    return ndl.dict_ndl(FILE_PATH_SIMPLE, ALPHA, BETAS, make_data_array=True)


@pytest.fixture(scope='module')
def result_continue_learning():
    events_simple = pd.read_csv(FILE_PATH_SIMPLE, sep="\t")
    part_1 = events_simple.head(CONTINUE_SPLIT_POINT)
    part_2 = events_simple.tail(len(events_simple) - CONTINUE_SPLIT_POINT)

    assert len(part_1) > 0  # pylint: disable=len-as-condition
    assert len(part_2) > 0  # pylint: disable=len-as-condition

    part_path_1 = os.path.join(TMP_PATH, "event_file_simple_1.tab.gz")
    part_path_2 = os.path.join(TMP_PATH, "event_file_simple_2.tab.gz")

    part_1.to_csv(part_path_1, header=True, index=None,
                  sep='\t', columns=["cues", "outcomes"],
                  compression='gzip')
    part_2.to_csv(part_path_2, header=True, index=None,
                  sep='\t', columns=["cues", "outcomes"],
                  compression='gzip')

    del events_simple, part_1, part_2

    result_part = ndl.ndl(part_path_1, ALPHA, BETAS)
    result = ndl.ndl(part_path_2, ALPHA, BETAS, weights=result_part)

    return result


def test_exceptions():
    with pytest.raises(ValueError) as e_info:
        ndl.ndl(FILE_PATH_SIMPLE, ALPHA, BETAS, method='threading', weights=1)
        assert e_info == 'weights need to be None or xarray.DataArray with method=threading'

    with pytest.raises(ValueError) as e_info:
        ndl.ndl(FILE_PATH_SIMPLE, ALPHA, BETAS, method='magic')
        assert e_info == 'method needs to be either "threading" or "openmp"'

    with pytest.raises(ValueError) as e_info:
        ndl.dict_ndl(FILE_PATH_SIMPLE, ALPHA, BETAS, weights=1)
        assert e_info == 'weights needs to be either defaultdict or None'

    with pytest.raises(ValueError) as e_info:
        ndl.dict_ndl(FILE_PATH_MULTIPLE_CUES, ALPHA, BETAS, remove_duplicates=None)
        assert e_info == 'cues or outcomes needs to be unique: cues "a a"; outcomes "A"; use remove_duplicates=True'

    with pytest.raises(ValueError) as e_info:
        ndl.ndl(FILE_PATH_SIMPLE, ALPHA, BETAS, method='threading',
                n_outcomes_per_job=-1)
        assert e_info == "'n_outcomes_per_job' must be larger then one"

    with pytest.raises(ValueError) as e_info:
        ndl.dict_ndl(FILE_PATH_SIMPLE, ALPHA, BETAS, make_data_array="magic")
        assert e_info == "make_data_array must be True or False"

    with pytest.raises(ValueError) as e_info:
        ndl.dict_ndl(FILE_PATH_SIMPLE, ALPHA, BETAS, remove_duplicates="magic")
        assert e_info == "remove_duplicates must be None, True or False"

    with pytest.raises(ValueError) as e_info:
        ndl.ndl(FILE_PATH_SIMPLE, ALPHA, BETAS, method='threading', remove_duplicates="magic")
        assert e_info == "remove_duplicates must be None, True or False"

    with pytest.raises(FileNotFoundError, match="No such file or directory") as e_info:
        ndl.ndl(FILE_PATH_SIMPLE, ALPHA, BETAS, method='threading', temporary_directory="./magic")

    with pytest.raises(ValueError, match="events_per_file has to be larger than 1") as e_info:
        ndl.ndl(FILE_PATH_SIMPLE, ALPHA, BETAS, method='threading', events_per_temporary_file=1)

    with pytest.raises(AttributeError, match="weights does not have attributes "
                       "and no attrs argument is given.") as e_info:
        ndl.data_array(dict())

#    # Test usually exeeds memory limit; It demands ~32GB of RAM.
#    with pytest.raises(ValueError, match="Neither number of cues nor outcomes "
#                       "shall exceed 4294967295 for now. See "
#                       "https://github.com/quantling/pyndl/issues/169") as e_info:
#        ndl.ndl(FILE_PATH_SIMPLE, ALPHA, BETAS,
#                weights=xr.DataArray(np.zeros(shape=(4294967295 + 1, 1))))


def test_generator_learning():
    events = io.events_from_file(FILE_PATH_SIMPLE)
    result_ndl_gen = ndl.ndl(events, ALPHA, BETAS, method='threading')
    result_ndl = ndl.ndl(FILE_PATH_SIMPLE, ALPHA, BETAS, method='threading')

    unequal, unequal_ratio = compare_arrays(FILE_PATH_SIMPLE,
                                            result_ndl_gen,
                                            result_ndl)
    print(result_ndl_gen)
    print('%.2f ratio unequal' % unequal_ratio)
    assert len(unequal) == 0  # pylint: disable=len-as-condition


def test_data_array_cast():
    result_ndl = ndl.ndl(FILE_PATH_SIMPLE, ALPHA, BETAS, method='threading')
    casted_result = ndl.data_array(result_ndl)
    assert isinstance(casted_result, xr.DataArray) and (result_ndl == casted_result).all()


def test_continue_learning_dict():
    events_simple = pd.read_csv(FILE_PATH_SIMPLE, sep="\t")
    part_1 = events_simple.head(CONTINUE_SPLIT_POINT)
    part_2 = events_simple.tail(len(events_simple) - CONTINUE_SPLIT_POINT)

    assert len(part_1) > 0  # pylint: disable=len-as-condition
    assert len(part_2) > 0  # pylint: disable=len-as-condition

    part_path_1 = os.path.join(TMP_PATH, "event_file_simple_1.tab.gz")
    part_path_2 = os.path.join(TMP_PATH, "event_file_simple_2.tab.gz")

    part_1.to_csv(part_path_1, header=True, index=None,
                  sep='\t', columns=["cues", "outcomes"],
                  compression='gzip')
    part_2.to_csv(part_path_2, header=True, index=None,
                  sep='\t', columns=["cues", "outcomes"],
                  compression='gzip')

    del events_simple, part_1, part_2

    result_part = ndl.dict_ndl(part_path_1,
                               ALPHA, BETAS)
    result_part_copy = copy.deepcopy(result_part)

    result_inplace = ndl.dict_ndl(part_path_2, ALPHA, BETAS,
                                  weights=result_part, inplace=True)

    assert result_part is result_inplace
    assert result_part != result_part_copy

    result_part = ndl.dict_ndl(part_path_1,
                               ALPHA, BETAS)

    result = ndl.dict_ndl(part_path_2,
                          ALPHA, BETAS, weights=result_part)

    assert result_part != result


def test_continue_learning_dict_ndl_data_array(result_dict_ndl, result_dict_ndl_data_array):
    continue_from_dict = ndl.dict_ndl(FILE_PATH_SIMPLE, ALPHA, BETAS,
                                      weights=result_dict_ndl)
    continue_from_data_array = ndl.dict_ndl(FILE_PATH_SIMPLE, ALPHA, BETAS,
                                            weights=result_dict_ndl_data_array)
    unequal, unequal_ratio = compare_arrays(FILE_PATH_SIMPLE,
                                            continue_from_dict,
                                            continue_from_data_array)
    print(continue_from_data_array)
    print('%.2f ratio unequal' % unequal_ratio)
    assert len(unequal) == 0  # pylint: disable=len-as-condition


@pytest.mark.nolinux
def test_continue_learning(result_continue_learning, result_ndl_openmp):
    assert result_continue_learning.shape == result_ndl_openmp.shape

    assert set(result_continue_learning.coords["outcomes"].values) == set(result_ndl_openmp.coords["outcomes"].values)

    assert set(result_continue_learning.coords["cues"].values) == set(result_ndl_openmp.coords["cues"].values)

    unequal, unequal_ratio = compare_arrays(FILE_PATH_SIMPLE,
                                            result_continue_learning,
                                            result_ndl_openmp)
    print('%.2f ratio unequal' % unequal_ratio)
    assert len(unequal) == 0  # pylint: disable=len-as-condition


@pytest.mark.nolinux
def test_save_to_netcdf4(result_ndl_openmp):
    weights = result_ndl_openmp.copy()  # avoids changing shared test data
    path = os.path.join(TMP_PATH, "weights.nc")
    weights.to_netcdf(path)
    weights_read = xr.open_dataarray(path)
    # does not preserves the order of the OrderedDict
    for key, value in weights.attrs.items():
        assert value == weights_read.attrs[key]
    weights_continued = ndl.ndl(FILE_PATH_SIMPLE, ALPHA, BETAS, method='openmp', weights=weights)
    path_continued = os.path.join(TMP_PATH, "weights_continued.nc")
    weights_continued.to_netcdf(path_continued)
    weights_continued_read = xr.open_dataarray(path_continued)
    for key, value in weights_continued.attrs.items():
        assert value == weights_continued_read.attrs[key]


@pytest.mark.nolinux
def test_return_values(result_dict_ndl, result_dict_ndl_data_array, result_ndl_threading, result_ndl_openmp):
    # dict_ndl
    assert isinstance(result_dict_ndl, defaultdict)
    assert isinstance(result_dict_ndl_data_array, xr.DataArray)
    # openmp
    assert isinstance(result_ndl_openmp, xr.DataArray)
    # threading
    assert isinstance(result_ndl_threading, xr.DataArray)


@pytest.mark.nolinux
def test_provide_temporary_directory():
    with tempfile.TemporaryDirectory(dir=TMP_PATH) as temporary_directory:
        ndl.ndl(FILE_PATH_SIMPLE, ALPHA, BETAS, temporary_directory=temporary_directory)


# Test internal consistency

def test_dict_ndl_vs_ndl_threading(result_dict_ndl, result_ndl_threading):
    unequal, unequal_ratio = compare_arrays(FILE_PATH_SIMPLE, result_dict_ndl,
                                            result_ndl_threading)
    print('%.2f ratio unequal' % unequal_ratio)
    assert len(unequal) == 0  # pylint: disable=len-as-condition


def test_dict_ndl_vs_dict_ndl_generator(result_dict_ndl, result_dict_ndl_generator):
    unequal, unequal_ratio = compare_arrays(FILE_PATH_SIMPLE, result_dict_ndl,
                                            result_dict_ndl_generator)
    print('%.2f ratio unequal' % unequal_ratio)
    assert len(unequal) == 0  # pylint: disable=len-as-condition


def test_dict_ndl_data_array_vs_ndl_threading(result_ndl_threading):
    result_dict_ndl = ndl.dict_ndl(FILE_PATH_SIMPLE, ALPHA, BETAS, make_data_array=True)

    unequal, unequal_ratio = compare_arrays(FILE_PATH_SIMPLE, result_dict_ndl,
                                            result_ndl_threading)
    print('%.2f ratio unequal' % unequal_ratio)
    assert len(unequal) == 0  # pylint: disable=len-as-condition


def test_ordering_of_temporary_event_files(result_dict_ndl):
    result_ndl = ndl.ndl(FILE_PATH_SIMPLE, ALPHA, BETAS, method='threading',
                         events_per_temporary_file=2)

    unequal, unequal_ratio = compare_arrays(FILE_PATH_SIMPLE, result_dict_ndl,
                                            result_ndl)
    print('%.2f ratio unequal' % unequal_ratio)
    assert len(unequal) == 0


def test_multiple_cues_dict_ndl_vs_ndl_threading():
    result_dict_ndl = ndl.dict_ndl(FILE_PATH_MULTIPLE_CUES, ALPHA, BETAS, remove_duplicates=True)
    result_ndl_threading = ndl.ndl(FILE_PATH_MULTIPLE_CUES, ALPHA, BETAS, remove_duplicates=True, method='threading')

    unequal, unequal_ratio = compare_arrays(FILE_PATH_MULTIPLE_CUES, result_dict_ndl,
                                            result_ndl_threading)
    print('%.2f ratio unequal' % unequal_ratio)
    assert len(unequal) == 0  # pylint: disable=len-as-condition


@pytest.mark.nolinux
def test_dict_ndl_vs_ndl_openmp(result_dict_ndl, result_ndl_openmp):
    result_dict_ndl = ndl.dict_ndl(FILE_PATH_SIMPLE, ALPHA, BETAS)
    unequal, unequal_ratio = compare_arrays(FILE_PATH_SIMPLE, result_dict_ndl,
                                            result_ndl_openmp)
    print('%.2f ratio unequal' % unequal_ratio)
    assert len(unequal) == 0  # pylint: disable=len-as-condition


@pytest.mark.nolinux
def test_meta_data(result_dict_ndl, result_dict_ndl_data_array, result_ndl_openmp, result_ndl_threading):
    attributes = {'cython', 'cpu_time', 'hostname', 'xarray', 'wall_time',
                  'event_path', 'number_events', 'username', 'method', 'date', 'numpy',
                  'betas', 'lambda', 'pyndl', 'alpha', 'pandas', 'method',
                  'function'}
    results = [result_dict_ndl, result_dict_ndl_data_array, result_ndl_threading, result_ndl_openmp]
    for result in results:
        assert set(result.attrs.keys()) == attributes

    assert int(result_dict_ndl_data_array.attrs['number_events']) > 0
    assert len(set(
        [result.attrs['number_events'].strip()
         for result in results]
    )) == 1


# Test against external ndl2 results
def test_compare_weights_ndl2(result_dict_ndl):
    """
    Checks whether the output of the R learner implemented in ndl2 and the
    python implementation of dict_ndl is equal.

    R code to generate the results::

        library(ndl2)
        learner <- learnWeightsTabular('event_file_simple.tab.gz', alpha=0.1, beta=0.1, lambda=1.0)
        wm <- learner$getWeights()
        wm <- wm[order(rownames(wm)), order(colnames(wm))]
        write.csv(wm, 'weights_event_file_simple_ndl2.csv')

    """
    result_ndl2 = defaultdict(lambda: defaultdict(float))

    with open(REFERENCE_PATH, 'rt') as reference_file:
        first_line = reference_file.readline().strip()
        outcomes = first_line.split(',')[1:]
        outcomes = [outcome.strip('"') for outcome in outcomes]
        for line in reference_file:
            cue, *cue_weights = line.strip().split(',')
            cue = cue.strip('"')
            for ii, outcome in enumerate(outcomes):
                result_ndl2[outcome][cue] = float(cue_weights[ii])

    unequal, unequal_ratio = compare_arrays(FILE_PATH_SIMPLE, result_ndl2, result_dict_ndl)
    print(set(outcome for outcome, *_ in unequal))
    print('%.2f ratio unequal' % unequal_ratio)
    assert len(unequal) == 0  # pylint: disable=len-as-condition


def test_multiple_cues_dict_ndl_vs_ndl2():
    """
    Checks whether the output of the R learner implemented in ndl2 and the
    python implementation of dict_ndl is equal.

    R code to generate the results::

        library(ndl2)
        learner <- learnWeightsTabular('tests/resources/event_file_multiple_cues.tab.gz',
                                       alpha=0.1, beta=0.1, lambda=1.0, removeDuplicates=FALSE)
        wm <- learner$getWeights()
        wm <- wm[order(rownames(wm)), order(colnames(wm))]
        write.csv(wm, 'tests/reference/weights_event_file_multiple_cues_ndl2.csv')

    """
    result_ndl2 = defaultdict(lambda: defaultdict(float))

    with open(REFERENCE_PATH_MULTIPLE_CUES_NDL2, 'rt') as reference_file:
        first_line = reference_file.readline().strip()
        outcomes = first_line.split(',')[1:]
        outcomes = [outcome.strip('"') for outcome in outcomes]
        for line in reference_file:
            cue, *cue_weights = line.strip().split(',')
            cue = cue.strip('"')
            for ii, outcome in enumerate(outcomes):
                result_ndl2[outcome][cue] = float(cue_weights[ii])

    result_python = ndl.dict_ndl(FILE_PATH_MULTIPLE_CUES, ALPHA, BETAS, remove_duplicates=False)

    unequal, unequal_ratio = compare_arrays(FILE_PATH_MULTIPLE_CUES, result_ndl2, result_python)
    print(set(outcome for outcome, *_ in unequal))
    print('%.2f ratio unequal' % unequal_ratio)
    assert len(unequal) == 0  # pylint: disable=len-as-condition


def test_compare_weights_rescorla_vs_ndl2():
    """
    Checks whether the output of the R learner implemented in ndl2 and the
    python implementation of dict_ndl is equal.

    R code to generate the results::

        library(ndl2)
        learner <- learnWeightsTabular('tests/resources/event_file_simple.tab.gz', alpha=0.1, beta=0.1, lambda=1.0)
        wm <- learner$getWeights()
        wm <- wm[order(rownames(wm)), order(colnames(wm))]
        write.csv(wm, 'tests/reference/weights_event_file_simple_ndl2.csv')

    """
    result_ndl2 = defaultdict(lambda: defaultdict(float))

    with open(REFERENCE_PATH, 'rt') as reference_file:
        first_line = reference_file.readline().strip()
        outcomes = first_line.split(',')[1:]
        outcomes = [outcome.strip('"') for outcome in outcomes]
        for line in reference_file:
            cue, *cue_weights = line.strip().split(',')
            cue = cue.strip('"')
            for ii, outcome in enumerate(outcomes):
                result_ndl2[outcome][cue] = float(cue_weights[ii])

    result_rescorla = defaultdict(lambda: defaultdict(float))

    with open(REFERENCE_PATH_NDL2, 'rt') as reference_file:
        first_line = reference_file.readline().strip()
        outcomes = first_line.split(',')[1:]
        outcomes = [outcome.strip('"') for outcome in outcomes]
        for line in reference_file:
            cue, *cue_weights = line.strip().split(',')
            cue = cue.strip('"')
            for ii, outcome in enumerate(outcomes):
                result_rescorla[outcome][cue] = float(cue_weights[ii])

    unequal, unequal_ratio = compare_arrays(FILE_PATH_SIMPLE, result_ndl2, result_rescorla)
    print('%.2f ratio unequal' % unequal_ratio)
    assert len(unequal) == 0  # pylint: disable=len-as-condition


@pytest.mark.runslow
def test_compare_time_dict_inplace_parallel_thread():
    file_path = os.path.join(TEST_ROOT, 'resources/event_file_many_cues.tab.gz')

    result_dict_ndl, duration_not_parallel = clock(ndl.dict_ndl, (file_path, ALPHA, BETAS, LAMBDA_))

    result_thread_ndl, duration_parallel = clock(ndl.ndl,
                                                 (file_path, ALPHA, BETAS, LAMBDA_),
                                                 n_jobs=4, method='threading')

    assert len(result_dict_ndl) == len(result_thread_ndl)

    unequal, unequal_ratio = compare_arrays(file_path, result_thread_ndl, result_dict_ndl)
    print('%.2f ratio unequal' % unequal_ratio)
    assert len(unequal) == 0  # pylint: disable=len-as-condition

    print('parallel: %.3e  dict: %.3e' % (duration_parallel, duration_not_parallel))
    assert duration_parallel < duration_not_parallel


def test_slice_list():
    lst = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]

    res = ndl.slice_list(lst, 2)
    assert res == [[0, 1], [2, 3], [4, 5], [6, 7], [8, 9]]

    res2 = ndl.slice_list(lst, 3)
    assert res2 == [[0, 1, 2], [3, 4, 5], [6, 7, 8], [9]]


def clock(func, args, **kwargs):
    start = time.time()
    result = func(*args, **kwargs)
    stop = time.time()

    duration = stop - start

    return result, duration


def compare_arrays(file_path, arr1, arr2):
    _, cues, outcomes = count.cues_outcomes(file_path)
    cue_map, outcome_map, _ = generate_mapping(file_path)

    unequal = list()

    for outcome in outcomes:
        for cue in cues:
            values = list()
            for array in (arr1, arr2):
                if isinstance(array, np.ndarray):
                    outcome_index = outcome_map[outcome]
                    cue_index = cue_map[cue]
                    values.append(array[outcome_index][cue_index])
                elif isinstance(array, xr.DataArray):
                    values.append(array.loc[{'outcomes': outcome, 'cues': cue}].values)
                elif isinstance(array, pd.DataFrame):
                    values.append(array.loc[outcome][cue])
                else:
                    values.append(array[outcome][cue])

            value1, value2 = values  # pylint: disable=unbalanced-tuple-unpacking
            if not np.isclose(value1, value2, rtol=1e-02, atol=1e-05):
                unequal.append((outcome, cue, value1, value2))

    unequal_ratio = len(unequal) / (len(outcomes) * len(cues))
    return (unequal, unequal_ratio)


def generate_mapping(event_path):
    _, cues, outcomes = count.cues_outcomes(event_path)
    all_cues = list(cues.keys())
    all_outcomes = list(outcomes.keys())
    cue_map = OrderedDict(((cue, ii) for ii, cue in enumerate(all_cues)))
    outcome_map = OrderedDict(((outcome, ii) for ii, outcome in enumerate(all_outcomes)))

    return (cue_map, outcome_map, all_outcomes)



#!/usr/bin/env python3

# pylint: disable=C0111

import os

from pyndl import count

TEST_ROOT = os.path.dirname(__file__)
EVENT_RESOURCE_FILE = os.path.join(TEST_ROOT, "resources/event_file_trigrams_to_word.tab.gz")
CORPUS_RESOURCE_FILE = os.path.join(TEST_ROOT, "resources/corpus.txt")


def test_cues_outcomes():
    n_events, cues, outcomes = count.cues_outcomes(EVENT_RESOURCE_FILE)
    n_events3, cues3, outcomes3 = count.cues_outcomes(EVENT_RESOURCE_FILE,
                                                      n_jobs=6,
                                                      verbose=True)
    assert n_events == 2772
    assert n_events == n_events3
    assert cues == cues3
    assert outcomes == outcomes3


def test_words_symbols():
    words, symbols = count.words_symbols(CORPUS_RESOURCE_FILE)
    words3, symbols3 = count.words_symbols(CORPUS_RESOURCE_FILE,
                                           n_jobs=3,
                                           verbose=True)
    assert words == words3
    assert symbols == symbols3


def test_save_load():
    file_name = os.path.join(TEST_ROOT, "temp/cues.tab")
    _, cues, _ = count.cues_outcomes(EVENT_RESOURCE_FILE)
    count.save_counter(cues, file_name)
    cues_loaded = count.load_counter(file_name)
    assert cues == cues_loaded
    os.remove(file_name)



#!/usr/bin/env python3

# pylint: disable=C0111

import os
import gzip

import pandas as pd

from pyndl import io

TEST_ROOT = os.path.join(os.path.pardir, os.path.dirname(__file__))
FILE_PATH_SIMPLE = os.path.join(TEST_ROOT, "resources/event_file_simple.tab.gz")
TMP_SAVE_PATH = os.path.join(TEST_ROOT, "temp/event_file_simple_copy.tab.gz")
TMP_SAVE_PATH_UNCOMPRESSED = os.path.join(TEST_ROOT, "temp/event_file_simple_copy.tab")


def test_uncompressed():
    data_frame = pd.read_csv(FILE_PATH_SIMPLE, sep="\t")
    ref_events = list(io.events_from_dataframe(data_frame))
    io.events_to_file(ref_events, TMP_SAVE_PATH_UNCOMPRESSED, compression=None)
    events = io.events_from_file(TMP_SAVE_PATH_UNCOMPRESSED, compression=None)

    unequal, unequal_ratio = compare_events(events, ref_events)

    print('%.2f ratio unequal' % unequal_ratio)
    assert len(unequal) == 0  # pylint: disable=len-as-condition

    os.remove(TMP_SAVE_PATH_UNCOMPRESSED)


def test_events_from_dataframe():
    data_frame = pd.read_csv(FILE_PATH_SIMPLE, sep="\t")
    events = io.events_from_dataframe(data_frame)
    ref_events = io.events_from_file(FILE_PATH_SIMPLE)

    unequal, unequal_ratio = compare_events(events, ref_events)

    print('%.2f ratio unequal' % unequal_ratio)
    assert len(unequal) == 0  # pylint: disable=len-as-condition


def test_events_from_list():
    data_frame = pd.read_csv(FILE_PATH_SIMPLE, sep="\t")
    event_list = [(cues, outcomes) for cues, outcomes in zip(data_frame['cues'],
                                                             data_frame['outcomes'])]

    events = io.events_from_list(event_list)

    ref_events = io.events_from_file(FILE_PATH_SIMPLE)

    unequal, unequal_ratio = compare_events(events, ref_events)

    print('%.2f ratio unequal' % unequal_ratio)
    assert len(unequal) == 0  # pylint: disable=len-as-condition


def test_events_to_file():
    data_frame = pd.read_csv(FILE_PATH_SIMPLE, sep="\t")
    io.events_to_file(data_frame, TMP_SAVE_PATH)

    with gzip.open(FILE_PATH_SIMPLE, 'rt') as ref_file:
        with gzip.open(TMP_SAVE_PATH, 'rt') as event_file:

            for line, ref_line in zip(event_file, ref_file):
                assert line == ref_line

    os.remove(TMP_SAVE_PATH)


def test_event_generator_to_file():
    events = io.events_from_file(FILE_PATH_SIMPLE)
    io.events_to_file(events, TMP_SAVE_PATH)

    with gzip.open(FILE_PATH_SIMPLE, 'rt') as ref_file:
        with gzip.open(TMP_SAVE_PATH, 'rt') as event_file:

            for line, ref_line in zip(event_file, ref_file):
                assert line == ref_line

    os.remove(TMP_SAVE_PATH)


def compare_events(events1, events2):
    events1, events2 = list(events1), list(events2)

    assert len(events1) == len(events2)

    unequal = list()

    for (cues1, outcomes1), (cues2, outcomes2) in zip(events1, events2):
        if sorted(cues1) != sorted(cues2) or sorted(outcomes1) != sorted(outcomes2):
            unequal.append(((cues1, outcomes1), (cues2, outcomes2)))

    unequal_ratio = len(unequal) / len(events1)
    return (unequal, unequal_ratio)



import numpy as np
cimport numpy as np

ctypedef np.float64_t dtype_t
cimport cython
from cython.parallel cimport parallel, prange


@cython.cdivision(True)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
def correlation(semantics_py,
                activations_py,
                np.ndarray[dtype_t, ndim=1] semantics_means,
                np.ndarray[dtype_t, ndim=1] semantics_stds,
                np.ndarray[dtype_t, ndim=1] activations_means,
                np.ndarray[dtype_t, ndim=1] activations_stds,
                *,
                unsigned int n_jobs=30,
                unsigned int chunksize=10):
    """
    The following formula is implemented here.

    .. math::

        r_{xy} = \frac{ \sum x_{i}y_{i}-n{\bar {x}}{\bar{y}} }{ (n-1)s_{x}s_{y} }


    """

    cdef int n_outcomes = semantics_py.shape[1]
    cdef int n_vec_dims, n_events = 0
    n_vec_dims, n_events = activations_py.shape
    cdef np.ndarray[dtype_t, ndim=2] correlations = np.zeros((n_outcomes, n_events))
    cdef int ii, jj, kk = 0
    cdef dtype_t scalar_prod, nominator, denominator = 0.0
    cdef dtype_t n_vec_dims_float = <dtype_t> n_vec_dims

    cdef np.ndarray[dtype_t, ndim=2] semantics = semantics_py
    cdef np.ndarray[dtype_t, ndim=2] activations = activations_py

    with nogil, parallel(num_threads=n_jobs):
        for ii in prange(n_events, schedule="dynamic", chunksize=chunksize):
            for jj in range(n_outcomes):
                scalar_prod = 0.0
                for kk in range(n_vec_dims):
                    scalar_prod = scalar_prod + semantics[kk, jj] * activations[kk, ii]
                nominator = scalar_prod - n_vec_dims_float * semantics_means[jj] * activations_means[ii]
                denominator = (n_vec_dims_float - 1.0) * semantics_stds[jj] * activations_stds[ii]

                correlations[jj, ii] = nominator / denominator

    return correlations




import numpy as np
import math
from libc.stdlib cimport abort, malloc, free
from libc.stdio cimport fopen, fread, fclose, FILE

from error_codes cimport ErrorCode, NO_ERROR, MAGIC_NUMBER_DOES_NOT_MATCH, VERSION_NUMBER_DOES_NOT_MATCH, INITIAL_ERROR_CODE, ERROR_CODES


cdef unsigned int MAGIC_NUMBER = 14159265
cdef unsigned int CURRENT_VERSION_WITH_FREQ = 215
cdef unsigned int CURRENT_VERSION = 2048 + 215


# run two sanity checks while loading the extension
# 1. check
if sizeof(unsigned int) != 4:
    raise ImportError('unsigned int needs to be 4 bytes not %i bytes' % sizeof(unsigned int))

# 2. check
# integer overflow in uni-dimensional index
cdef unsigned long long test_index
cdef unsigned int test_cue_index, test_mm, test_outcome_index
test_cue_index = test_mm = test_outcome_index = 4294967295

test_index = test_mm
test_index *= test_outcome_index
test_index += test_cue_index
assert test_index == 18446744069414584320


cdef inline void read_next_int(void *data, FILE *binary_file) nogil:
    fread(data, 4, 1, binary_file) # little endian


cdef extern from "stdio.h":
    #FILE * fopen ( const char * filename, const char * mode )
    FILE *fopen(const char *, const char *) nogil
    #int fclose ( FILE * stream )
    int fclose(FILE *) nogil
    #size_t fread ( void *ptr, size_t size, size_t count, FILE *stream );
    size_t fread (void *, size_t, size_t, FILE *) nogil


def learn_inplace_binary_to_binary(binary_file_paths,
                  dtype_t alpha, dtype_t beta1,
                  dtype_t beta2, dtype_t lambda_,
                  np.ndarray[dtype_t, ndim=2] weights,
                  np.ndarray[unsigned int, ndim=1] all_outcomes):

    cdef unsigned int n_all_cues = weights.shape[1]  # number of cues == columns
    cdef unsigned int* all_outcomes_ptr = <unsigned int *> all_outcomes.data
    cdef unsigned int length_all_outcomes = all_outcomes.shape[0]
    cdef char* fname
    cdef unsigned int start_val, end_val
    cdef ErrorCode error = INITIAL_ERROR_CODE

  #  cdef String
    # weights muss contigousarray sein und mode=c, siehe:
    #cdef np.ndarray[np.uint32_t, ndim=3, mode = 'c'] np_buff = np.ascontiguousarray(im, dtype = np.uint32)
    cdef dtype_t* weights_ptr = <dtype_t *> weights.data # ueberlegen ob [][] oder ** oder [] oder *

    for binary_file_path in binary_file_paths: #
        filename_byte_string = binary_file_path.encode("UTF-8")
        fname = filename_byte_string

        with nogil:
            error = learn_inplace_binary_to_binary_ptr(fname, alpha, beta1, beta2, lambda_,
                              weights_ptr, n_all_cues, all_outcomes_ptr, 0,
                              length_all_outcomes)
            if error != NO_ERROR:
                break

    if (error != NO_ERROR):
        raise IOError(f'binary files does not have proper format, error code {error}\n{ERROR_CODES}')



cdef int is_element_of(unsigned int elem, unsigned int* arr, unsigned int size) nogil:
    cdef unsigned int ii
    for ii in range(size):
        if arr[ii] == elem:
            return True
    return False


# ggf exception zurückgeben
cdef ErrorCode learn_inplace_binary_to_binary_ptr(char* binary_file_path,
                        dtype_t alpha, dtype_t beta1,
                        dtype_t beta2, dtype_t lambda_,
                        dtype_t* weights,
                        unsigned int n_all_cues,
                        unsigned int* all_outcome_indices,
                        unsigned int start,
                        unsigned int end) nogil:


    cdef unsigned int number_of_events, number_of_cues, number_of_outcomes
    cdef dtype_t association_strength, update
    cdef unsigned int magic_number, version, ii, jj, event, appearance
    cdef unsigned long long index
    cdef unsigned int* cue_indices
    cdef unsigned int* outcome_indices
    cdef unsigned int max_number_of_cues = 1024
    cdef unsigned int max_number_of_outcomes = 1024

    cdef FILE* binary_file
    binary_file = fopen(binary_file_path, "rb")

    read_next_int(&magic_number, binary_file)
    if not magic_number == MAGIC_NUMBER:
        fclose(binary_file)
        return MAGIC_NUMBER_DOES_NOT_MATCH
    read_next_int(&version, binary_file)
    if version == CURRENT_VERSION:
        pass
    else:
        fclose(binary_file)
        return VERSION_NUMBER_DOES_NOT_MATCH

    # preallocate memory
    cue_indices = <unsigned int *> malloc(sizeof(unsigned int) * max_number_of_cues)
    outcome_indices = <unsigned int *> malloc(sizeof(unsigned int) * max_number_of_outcomes)

    read_next_int(&number_of_events, binary_file)

    for event in range(number_of_events):
        # cues
        read_next_int(&number_of_cues, binary_file)
        if number_of_cues > max_number_of_cues:
            max_number_of_cues = number_of_cues
            free(cue_indices)
            cue_indices = <unsigned int *> malloc(sizeof(unsigned int) * max_number_of_cues)
        fread(cue_indices, 4, number_of_cues, binary_file)

        # outcomes
        read_next_int(&number_of_outcomes, binary_file)
        if number_of_outcomes > max_number_of_outcomes:
            max_number_of_outcomes = number_of_outcomes
            free(outcome_indices)
            outcome_indices = <unsigned int *> malloc(sizeof(unsigned int) * max_number_of_outcomes)
        fread(outcome_indices, 4, number_of_outcomes, binary_file)

        # learn
        for ii in range(start, end):
            association_strength = 0.0
            for jj in range(number_of_cues):
                # this overflows:
                #index = cue_indices[jj] + mm * all_outcome_indices[ii]
                index = n_all_cues  # implicit cast to unsigned long long
                index *=  all_outcome_indices[ii]  # this can't overflow anymore
                index += cue_indices[jj]  # this can't overflow anymore
                # worst case: 4294967295 * 4294967295 + 4294967295 == 18446744069414584320 < 18446744073709551615
                association_strength += weights[index]
            if is_element_of(all_outcome_indices[ii], outcome_indices, number_of_outcomes):
                update = beta1 * (lambda_ - association_strength)
            else:
                update = beta2 * (0.0 - association_strength)
            for jj in range(number_of_cues):
                index = n_all_cues  # implicit cast to unsigned long long
                index *=  all_outcome_indices[ii]  # this can't overflow anymore
                index += cue_indices[jj]  # this can't overflow anymore
                weights[index] += alpha * update

    fclose(binary_file)
    free(cue_indices)
    free(outcome_indices)
    return NO_ERROR


# ggf exception zurückgeben
cdef ErrorCode learn_inplace_binary_to_real_ptr(char* binary_file_path,
                        dtype_t eta,
                        dtype_t* outcome_vectors,
                        dtype_t* weights,
                        unsigned int n_all_cues,
                        unsigned int n_outcome_vector_dimensions,
                        unsigned int start,
                        unsigned int end) nogil:


    cdef unsigned int number_of_events, number_of_cues, number_of_outcomes
    cdef dtype_t association_strength, update, summed_outcome_vector_value
    cdef unsigned int magic_number, version, outcome_vec_dim_ii, jj, kk, event
    cdef unsigned long long index
    cdef unsigned int* cue_indices
    cdef unsigned int* outcome_indices
    cdef unsigned int max_number_of_cues = 1024
    cdef unsigned int max_number_of_outcomes = 1024

    cdef FILE* binary_file
    binary_file = fopen(binary_file_path, "rb")

    read_next_int(&magic_number, binary_file)
    if not magic_number == MAGIC_NUMBER:
        fclose(binary_file)
        return MAGIC_NUMBER_DOES_NOT_MATCH
    read_next_int(&version, binary_file)
    if version == CURRENT_VERSION:
        pass
    else:
        fclose(binary_file)
        return VERSION_NUMBER_DOES_NOT_MATCH

    # preallocate memory
    cue_indices = <unsigned int *> malloc(sizeof(unsigned int) * max_number_of_cues)
    outcome_indices = <unsigned int *> malloc(sizeof(unsigned int) * max_number_of_outcomes)

    read_next_int(&number_of_events, binary_file)

    for event in range(number_of_events):
        # cues
        read_next_int(&number_of_cues, binary_file)
        if number_of_cues > max_number_of_cues:
            max_number_of_cues = number_of_cues
            free(cue_indices)
            cue_indices = <unsigned int *> malloc(sizeof(unsigned int) * max_number_of_cues)
        fread(cue_indices, 4, number_of_cues, binary_file)

        # outcomes
        read_next_int(&number_of_outcomes, binary_file)
        if number_of_outcomes > max_number_of_outcomes:
            max_number_of_outcomes = number_of_outcomes
            free(outcome_indices)
            outcome_indices = <unsigned int *> malloc(sizeof(unsigned int) * max_number_of_outcomes)
        fread(outcome_indices, 4, number_of_outcomes, binary_file)

        # learn
        for outcome_vec_dim_ii in range(start, end):
            association_strength = 0.0
            for jj in range(number_of_cues):
                # this overflows:
                #index = cue_indices[jj] + n_all_cues * outcome_vec_dim_ii
                index = n_all_cues  # implicit cast to unsigned long long
                index *= outcome_vec_dim_ii  # this can't overflow anymore
                index += cue_indices[jj]  # this can't overflow anymore
                # worst case: 4294967295 * 4294967295 + 4294967295 == 18446744069414584320 < 18446744073709551615
                association_strength += weights[index]
            summed_outcome_vector_value = 0.0
            for kk in range(number_of_outcomes):
                index = n_outcome_vector_dimensions
                index *= outcome_indices[kk]
                index += outcome_vec_dim_ii
                summed_outcome_vector_value += outcome_vectors[index]
            update = eta * (summed_outcome_vector_value - association_strength)
            for jj in range(number_of_cues):
              index = n_all_cues  # implicit cast to unsigned long long
              index *=  outcome_vec_dim_ii  # this can't overflow anymore
              index += cue_indices[jj]  # this can't overflow anymore
              weights[index] += update

    fclose(binary_file)
    free(cue_indices)
    free(outcome_indices)
    return NO_ERROR


cdef ErrorCode learn_inplace_real_to_real_ptr(char* binary_file_path,
                        dtype_t eta,
                        dtype_t* cue_vectors,
                        dtype_t* outcome_vectors,
                        dtype_t* weights,
                        unsigned int n_cue_vector_dimensions,
                        unsigned int n_outcome_vector_dimensions,
                        unsigned int start,
                        unsigned int end) nogil:

    cdef unsigned int number_of_events, number_of_cues, number_of_outcomes
    cdef dtype_t association_strength, update, summed_cue_vector_value, summed_outcome_vector_value
    cdef unsigned int magic_number, version, ii, jj, kk, event, outcome_vec_dim_ii
    cdef unsigned long long index, index_cue, index_weight
    cdef unsigned int* cue_indices
    cdef unsigned int* outcome_indices
    cdef unsigned int max_number_of_cues = 1024
    cdef unsigned int max_number_of_outcomes = 1024

    cdef FILE* binary_file
    binary_file = fopen(binary_file_path, "rb")

    read_next_int(&magic_number, binary_file)
    if not magic_number == MAGIC_NUMBER:
        fclose(binary_file)
        return MAGIC_NUMBER_DOES_NOT_MATCH
    read_next_int(&version, binary_file)
    if version == CURRENT_VERSION:
        pass
    else:
        fclose(binary_file)
        return VERSION_NUMBER_DOES_NOT_MATCH

    # preallocate memory
    cue_indices = <unsigned int *> malloc(sizeof(unsigned int) * max_number_of_cues)
    outcome_indices = <unsigned int *> malloc(sizeof(unsigned int) * max_number_of_outcomes)

    read_next_int(&number_of_events, binary_file)

    for event in range(number_of_events):
        # cues
        read_next_int(&number_of_cues, binary_file)
        if number_of_cues > max_number_of_cues:
            max_number_of_cues = number_of_cues
            free(cue_indices)
            cue_indices = <unsigned int *> malloc(sizeof(unsigned int) * max_number_of_cues)
        fread(cue_indices, 4, number_of_cues, binary_file)

        # outcomes
        read_next_int(&number_of_outcomes, binary_file)
        if number_of_outcomes > max_number_of_outcomes:
            max_number_of_outcomes = number_of_outcomes
            free(outcome_indices)
            outcome_indices = <unsigned int *> malloc(sizeof(unsigned int) * max_number_of_outcomes)
        fread(outcome_indices, 4, number_of_outcomes, binary_file)

        # learn
        # start and end are refering to the vector dimensions of the outcome vector not to the outcomes
        for outcome_vec_dim_ii in range(start, end):
            ## W[ii, :] @ cue_vectors[jj] + W[ii, :] @ cue_vectors[jj]  + ...
            #association_strength = 0.0
            #for jj in range(number_of_cues):
            #    index_weight = n_cue_vector_dimensions
            #    index_weight *= outcome_vec_dim_ii
            #    index_weight += kk
            #    for kk in range(n_cue_vector_dimensions):
            #        index_cue = n_cue_vector_dimensions
            #        index_cue *= all_cue_indices[jj]
            #        index_cue += kk
            #        association_strength += cue_vectors[index_cue] * weights[index_weight]
            # = W[ii, :] @ (cue_vectors[jj] + cue_vectors[jj] + ...)
            association_strength = 0.0
            for kk in range(n_cue_vector_dimensions):
                summed_cue_vector_value = 0.0
                for jj in range(number_of_cues):
                    index_cue = n_cue_vector_dimensions
                    index_cue *= cue_indices[jj]
                    index_cue += kk
                    summed_cue_vector_value += cue_vectors[index_cue]
                index_weight = n_cue_vector_dimensions
                index_weight *= outcome_vec_dim_ii
                index_weight += kk
                association_strength += summed_cue_vector_value * weights[index_weight]

            summed_outcome_vector_value = 0.0
            for jj in range(number_of_outcomes):
                index = n_outcome_vector_dimensions
                index *= outcome_indices[jj]
                index += outcome_vec_dim_ii
                summed_outcome_vector_value += outcome_vectors[index]
            # update = prediction error in learning * learning rate
            update = eta * (summed_outcome_vector_value - association_strength)

            for kk in range(n_cue_vector_dimensions):
                summed_cue_vector_value = 0.0
                for jj in range(number_of_cues):
                    index_cue = n_cue_vector_dimensions
                    index_cue *= cue_indices[jj]
                    index_cue += kk
                    summed_cue_vector_value += cue_vectors[index_cue]
                index_weight = n_cue_vector_dimensions
                index_weight *= outcome_vec_dim_ii
                index_weight += kk
                weights[index_weight] += update * summed_cue_vector_value

    fclose(binary_file)
    free(cue_indices)
    free(outcome_indices)
    return NO_ERROR


cdef ErrorCode learn_inplace_real_to_binary_ptr(char* binary_file_path,
                        dtype_t beta1,
                        dtype_t beta2,
                        dtype_t lambda_,
                        dtype_t* cue_vectors,
                        dtype_t* weights,
                        unsigned int n_cue_vector_dimensions,
                        unsigned int start,
                        unsigned int end) nogil:

    cdef unsigned int number_of_events, number_of_cues, number_of_outcomes
    cdef dtype_t association_strength, update, summed_cue_vector_value
    cdef unsigned int magic_number, version, ii, jj, kk, event
    cdef unsigned long long index, index_cue, index_weight
    cdef unsigned int* cue_indices
    cdef unsigned int* outcome_indices
    cdef unsigned int max_number_of_cues = 1024
    cdef unsigned int max_number_of_outcomes = 1024

    cdef FILE* binary_file
    binary_file = fopen(binary_file_path, "rb")

    read_next_int(&magic_number, binary_file)
    if not magic_number == MAGIC_NUMBER:
        fclose(binary_file)
        return MAGIC_NUMBER_DOES_NOT_MATCH
    read_next_int(&version, binary_file)
    if version == CURRENT_VERSION:
        pass
    else:
        fclose(binary_file)
        return VERSION_NUMBER_DOES_NOT_MATCH

    # preallocate memory
    cue_indices = <unsigned int *> malloc(sizeof(unsigned int) * max_number_of_cues)
    outcome_indices = <unsigned int *> malloc(sizeof(unsigned int) * max_number_of_outcomes)

    read_next_int(&number_of_events, binary_file)

    for event in range(number_of_events):
        # cues
        read_next_int(&number_of_cues, binary_file)
        if number_of_cues > max_number_of_cues:
            max_number_of_cues = number_of_cues
            free(cue_indices)
            cue_indices = <unsigned int *> malloc(sizeof(unsigned int) * max_number_of_cues)
        fread(cue_indices, 4, number_of_cues, binary_file)

        # outcomes
        read_next_int(&number_of_outcomes, binary_file)
        if number_of_outcomes > max_number_of_outcomes:
            max_number_of_outcomes = number_of_outcomes
            free(outcome_indices)
            outcome_indices = <unsigned int *> malloc(sizeof(unsigned int) * max_number_of_outcomes)
        fread(outcome_indices, 4, number_of_outcomes, binary_file)

        # learn
        for ii in range(start, end):
            ## W[ii, :] @ cue_vectors[jj] + W[ii, :] @ cue_vectors[jj]  + ...
            #association_strength = 0.0
            #for jj in range(number_of_cues):
            #    index_weight = n_cue_vector_dimensions
            #    index_weight *= outcome_vec_dim_ii
            #    index_weight += kk
            #    for kk in range(n_cue_vector_dimensions):
            #        index_cue = n_cue_vector_dimensions
            #        index_cue *= all_cue_indices[jj]
            #        index_cue += kk
            #        association_strength += cue_vectors[index_cue] * weights[index_weight]
            # = W[ii, :] @ (cue_vectors[jj] + cue_vectors[jj] + ...)
            association_strength = 0.0
            for kk in range(n_cue_vector_dimensions):
                summed_cue_vector_value = 0.0
                for jj in range(number_of_cues):
                    index_cue = n_cue_vector_dimensions
                    index_cue *= cue_indices[jj]
                    index_cue += kk
                    summed_cue_vector_value += cue_vectors[index_cue]
                index_weight = n_cue_vector_dimensions
                index_weight *= ii
                index_weight += kk
                association_strength += summed_cue_vector_value * weights[index_weight]

            # update = prediction error in learning * learning rate
            if is_element_of(ii, outcome_indices, number_of_outcomes):
                update = beta1 * (lambda_ - association_strength)
            else:
                update = beta2 * (0.0 - association_strength)

            for kk in range(n_cue_vector_dimensions):
                summed_cue_vector_value = 0.0
                for jj in range(number_of_cues):
                    index_cue = n_cue_vector_dimensions
                    index_cue *= cue_indices[jj]
                    index_cue += kk
                    summed_cue_vector_value += cue_vectors[index_cue]
                index_weight = n_cue_vector_dimensions
                index_weight *= ii
                index_weight += kk
                weights[index_weight] += update * summed_cue_vector_value

    fclose(binary_file)
    free(cue_indices)
    free(outcome_indices)
    return NO_ERROR
